from __future__ import annotations

import copy
import logging
import random
from typing import Optional

import igraph as ig
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pydantic import BaseModel

from napistu import sbml_dfs_core
from napistu import utils

from napistu.constants import DEFAULT_WT_TRANS
from napistu.constants import DEFINED_WEIGHT_TRANSFORMATION
from napistu.constants import MINI_SBO_FROM_NAME
from napistu.constants import MINI_SBO_TO_NAME
from napistu.constants import SBML_DFS
from napistu.constants import SBO_MODIFIER_NAMES
from napistu.constants import SCORE_CALIBRATION_POINTS_DICT
from napistu.constants import ENTITIES_W_DATA
from napistu.constants import SOURCE_VARS_DICT

from napistu.network.constants import CPR_GRAPH_NODES
from napistu.network.constants import CPR_GRAPH_EDGES
from napistu.network.constants import CPR_GRAPH_EDGE_DIRECTIONS
from napistu.network.constants import CPR_GRAPH_REQUIRED_EDGE_VARS
from napistu.network.constants import CPR_GRAPH_NODE_TYPES
from napistu.network.constants import CPR_GRAPH_TYPES
from napistu.network.constants import CPR_WEIGHTING_STRATEGIES
from napistu.network.constants import SBOTERM_NAMES
from napistu.network.constants import REGULATORY_GRAPH_HIERARCHY
from napistu.network.constants import SURROGATE_GRAPH_HIERARCHY
from napistu.network.constants import VALID_CPR_GRAPH_TYPES
from napistu.network.constants import VALID_WEIGHTING_STRATEGIES

logger = logging.getLogger(__name__)


def create_cpr_graph(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    reaction_graph_attrs: Optional[dict] = None,
    directed: bool = True,
    edge_reversed: bool = False,
    graph_type: str = CPR_GRAPH_TYPES.BIPARTITE,
    verbose: bool = False,
    custom_transformations: Optional[dict] = None,
) -> ig.Graph:
    """
    Create CPR Graph

    Create an igraph network from a mechanistic network using one of a set of graph_types.

    Parameters
    ----------
    sbml_dfs : SBML_dfs
        A model formed by aggregating pathways
    reaction_graph_attrs: dict
        Dictionary containing attributes to pull out of reaction_data and
        a weighting scheme for the graph
    directed : bool
        Should a directed (True) or undirected graph be made (False)
    edge_reversed : bool
        Should the directions of edges be reversed or not (False)
    graph_type : str
        Type of graph to create, valid values are:
            - bipartite: substrates and modifiers point to the reaction they drive, this reaction points to products
            - reguatory: non-enzymatic modifiers point to enzymes, enzymes point to substrates and products
            - surrogate: non-enzymatic modifiers -> substrates -> enzymes -> reaction -> products.
              In this representation enzymes are effective standing in for their reaction (eventhough the enzyme is
              not modified by a substrate per-se).
    verbose : bool
        Extra reporting
    custom_transformations : dict, optional
        Dictionary of custom transformation functions to use for attribute transformation.

    Returns:
    ----------
    An Igraph network
    """

    if reaction_graph_attrs is None:
        reaction_graph_attrs = {}

    if not isinstance(sbml_dfs, sbml_dfs_core.SBML_dfs):
        raise TypeError(
            f"sbml_dfs must be a sbml_dfs_core.SBML_dfs, but was {type(sbml_dfs)}"
        )

    if not isinstance(reaction_graph_attrs, dict):
        raise TypeError(
            f"reaction_graph_attrs must be a dict, but was {type(reaction_graph_attrs)}"
        )

    if not isinstance(directed, bool):
        raise TypeError(f"directed must be a bool, but was {type(directed)}")

    if not isinstance(edge_reversed, bool):
        raise TypeError(f"edge_reverse must be a bool, but was {type(edge_reversed)}")

    if not isinstance(graph_type, str):
        raise TypeError(f"graph_type must be a str, but was {type(verbose)}")

    if graph_type not in VALID_CPR_GRAPH_TYPES:
        raise ValueError(
            f"graph_type is not a valid value ({graph_type}), valid values are {','.join(VALID_CPR_GRAPH_TYPES)}"
        )

    if not isinstance(verbose, bool):
        raise TypeError(f"verbose must be a bool, but was {type(verbose)}")

    # fail fast in reaction_graph_attrs is not properly formatted
    for k in reaction_graph_attrs.keys():
        _validate_entity_attrs(reaction_graph_attrs[k])

    working_sbml_dfs = copy.deepcopy(sbml_dfs)
    reaction_species_counts = working_sbml_dfs.reaction_species.value_counts(
        SBML_DFS.R_ID
    )
    valid_reactions = reaction_species_counts[reaction_species_counts > 1].index
    # due to autoregulation reactions, and removal of cofactors some
    # reactions may have 1 (or even zero) species. drop these.

    n_dropped_reactions = working_sbml_dfs.reactions.shape[0] - len(valid_reactions)
    if n_dropped_reactions != 0:
        logger.info(
            f"Dropping {n_dropped_reactions} reactions with <= 1 reaction species "
            "these underspecified reactions may be due to either unrepresented "
            "autoregulation and/or removal of cofactors."
        )

        working_sbml_dfs.reactions = working_sbml_dfs.reactions[
            working_sbml_dfs.reactions.index.isin(valid_reactions)
        ]
        working_sbml_dfs.reaction_species = working_sbml_dfs.reaction_species[
            working_sbml_dfs.reaction_species[SBML_DFS.R_ID].isin(valid_reactions)
        ]

    logger.info(
        "Organizing all network nodes (compartmentalized species and reactions)"
    )

    network_nodes = list()
    network_nodes.append(
        working_sbml_dfs.compartmentalized_species.reset_index()[
            [SBML_DFS.SC_ID, SBML_DFS.SC_NAME]
        ]
        .rename(columns={SBML_DFS.SC_ID: "node_id", SBML_DFS.SC_NAME: "node_name"})
        .assign(node_type=CPR_GRAPH_NODE_TYPES.SPECIES)
    )
    network_nodes.append(
        working_sbml_dfs.reactions.reset_index()[[SBML_DFS.R_ID, SBML_DFS.R_NAME]]
        .rename(columns={SBML_DFS.R_ID: "node_id", SBML_DFS.R_NAME: "node_name"})
        .assign(node_type=CPR_GRAPH_NODE_TYPES.REACTION)
    )

    # rename nodes to name since it is treated specially
    network_nodes_df = pd.concat(network_nodes).rename(
        columns={"node_id": CPR_GRAPH_NODES.NAME}
    )

    logger.info(f"Formatting edges as a {graph_type} graph")

    if graph_type == CPR_GRAPH_TYPES.BIPARTITE:
        network_edges = _create_cpr_graph_bipartite(working_sbml_dfs)
    elif graph_type in [CPR_GRAPH_TYPES.REGULATORY, CPR_GRAPH_TYPES.SURROGATE]:
        # pass graph_type so that an appropriate tiered schema can be used.
        network_edges = _create_cpr_graph_tiered(working_sbml_dfs, graph_type)
    else:
        raise NotImplementedError("Invalid graph_type")

    logger.info("Adding reversibility and other meta-data from reactions_data")
    augmented_network_edges = _augment_network_edges(
        network_edges,
        working_sbml_dfs,
        reaction_graph_attrs,
        custom_transformations=custom_transformations,
    )

    logger.info(
        "Creating reverse reactions for reversible reactions on a directed graph"
    )
    if directed:
        directed_network_edges = pd.concat(
            [
                # assign forward edges
                augmented_network_edges.assign(
                    direction=CPR_GRAPH_EDGE_DIRECTIONS.FORWARD
                ),
                # create reverse edges for reversibile reactions
                _reverse_network_edges(augmented_network_edges),
            ]
        )
    else:
        directed_network_edges = augmented_network_edges.assign(
            direction=CPR_GRAPH_EDGE_DIRECTIONS.UNDIRECTED
        )

    # de-duplicate edges
    unique_edges = (
        directed_network_edges.groupby([CPR_GRAPH_EDGES.FROM, CPR_GRAPH_EDGES.TO])
        .first()
        .reset_index()
    )

    if unique_edges.shape[0] != directed_network_edges.shape[0]:
        logger.warning(
            f"{directed_network_edges.shape[0] - unique_edges.shape[0]} edges were dropped "
            "due to duplicated origin -> target relationiships, use verbose for "
            "more information"
        )

        if verbose:
            # report duplicated edges
            grouped_edges = directed_network_edges.groupby(
                [CPR_GRAPH_EDGES.FROM, CPR_GRAPH_EDGES.TO]
            )
            duplicated_edges = [
                grouped_edges.get_group(x)
                for x in grouped_edges.groups
                if grouped_edges.get_group(x).shape[0] > 1
            ]
            example_duplicates = pd.concat(
                random.sample(duplicated_edges, min(5, len(duplicated_edges)))
            )

            logger.warning(utils.style_df(example_duplicates, headers="keys"))

    # reverse edge directions if edge_reversed is True:

    if edge_reversed:
        rev_unique_edges_df = unique_edges.copy()
        rev_unique_edges_df[CPR_GRAPH_EDGES.FROM] = unique_edges[CPR_GRAPH_EDGES.TO]
        rev_unique_edges_df[CPR_GRAPH_EDGES.TO] = unique_edges[CPR_GRAPH_EDGES.FROM]
        rev_unique_edges_df[CPR_GRAPH_EDGES.SC_PARENTS] = unique_edges[
            CPR_GRAPH_EDGES.SC_CHILDREN
        ]
        rev_unique_edges_df[CPR_GRAPH_EDGES.SC_CHILDREN] = unique_edges[
            CPR_GRAPH_EDGES.SC_PARENTS
        ]
        rev_unique_edges_df[CPR_GRAPH_EDGES.STOICHOMETRY] = unique_edges[
            CPR_GRAPH_EDGES.STOICHOMETRY
        ] * (-1)

        rev_unique_edges_df[CPR_GRAPH_EDGES.DIRECTION] = unique_edges[
            CPR_GRAPH_EDGES.DIRECTION
        ].replace(
            {
                CPR_GRAPH_EDGE_DIRECTIONS.REVERSE: CPR_GRAPH_EDGE_DIRECTIONS.FORWARD,
                CPR_GRAPH_EDGE_DIRECTIONS.FORWARD: CPR_GRAPH_EDGE_DIRECTIONS.REVERSE,
            }
        )
    else:
        # unchanged if edge_reversed is False:
        rev_unique_edges_df = unique_edges

    # convert nodes and edgelist into an igraph network

    logger.info("Formatting cpr_graph output")
    cpr_graph = ig.Graph.DictList(
        vertices=network_nodes_df.to_dict("records"),
        edges=rev_unique_edges_df.to_dict("records"),
        directed=directed,
        vertex_name_attr=CPR_GRAPH_NODES.NAME,
        edge_foreign_keys=(CPR_GRAPH_EDGES.FROM, CPR_GRAPH_EDGES.TO),
    )

    return cpr_graph


def process_cpr_graph(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    reaction_graph_attrs: Optional[dict] = None,
    directed: bool = True,
    edge_reversed: bool = False,
    graph_type: str = CPR_GRAPH_TYPES.BIPARTITE,
    weighting_strategy: str = CPR_WEIGHTING_STRATEGIES.UNWEIGHTED,
    verbose: bool = False,
    custom_transformations: dict = None,
) -> ig.Graph:
    """
    Process Consensus Graph

    Setup an igraph network and then add weights and other maleable attributes.

    Args:
        sbml_dfs (SBML_dfs): A model formed by aggregating pathways
        reaction_graph_attrs (dict): Dictionary containing attributes to pull out of reaction_data and
            a weighting scheme for the graph
        directed (bool): Should a directed (True) or undirected graph be made (False)
        edge_reversed (bool): Should directions of edges be reversed (False)
        graph_type (str): Type of graph to create, valid values are:
            - bipartite: substrates and modifiers point to the reaction they drive, this reaction points to products
            - reguatory: non-enzymatic modifiers point to enzymes, enzymes point to substrates and products
        weighting_strategy (str) : a network weighting strategy with options:
            - unweighted: all weights (and upstream_weights for directed graphs) are set to 1.
            - topology: weight edges by the degree of the source nodes favoring nodes emerging from nodes
                with few connections.
            - mixed: transform edges with a quantitative score based on reaction_attrs; and set edges
                without quantitative score as a source-specific weight.
            - calibrated: transforme edges with a quantitative score based on reaction_attrs and combine them
                with topology scores to generate a consensus.
        verbose (bool): Extra reporting
        custom_transformations (dict, optional):
            Dictionary of custom transformation functions to use for attribute transformation.

    Returns:
        weighted_graph (ig.Graph): An Igraph network
    """

    if reaction_graph_attrs is None:
        reaction_graph_attrs = {}

    logging.info("Constructing network")
    cpr_graph = create_cpr_graph(
        sbml_dfs,
        reaction_graph_attrs,
        directed=directed,
        edge_reversed=edge_reversed,
        graph_type=graph_type,
        verbose=verbose,
        custom_transformations=custom_transformations,
    )

    if "reactions" in reaction_graph_attrs.keys():
        reaction_attrs = reaction_graph_attrs["reactions"]
    else:
        reaction_attrs = dict()

    logging.info(f"Adding edge weights with an {weighting_strategy} strategy")

    weighted_cpr_graph = add_graph_weights(
        cpr_graph=cpr_graph,
        reaction_attrs=reaction_attrs,
        weighting_strategy=weighting_strategy,
    )

    return weighted_cpr_graph


def pluck_entity_data(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    graph_attrs: dict[str, dict],
    data_type: str,
    custom_transformations: Optional[dict[str, callable]] = None,
) -> pd.DataFrame | None:
    """
    Pluck Entity Attributes

    Pull species or reaction attributes out of an sbml_dfs based on a set of
      tables and variables to look for.

    Parameters:
    sbml_dfs: sbml_dfs_core.SBML_dfs
        A mechanistic model
    graph_attrs: dict
        A dictionary of species/reaction attributes to pull out. If the requested
        data_type ("species" or "reactions") is not present as a key, or if the value
        is an empty dict, this function will return None (no error).
    data_type: str
        "species" or "reactions" to pull out species_data or reactions_data
    custom_transformations: dict[str, callable], optional
        A dictionary mapping transformation names to functions. If provided, these
        will be checked before built-in transformations. Example:
            custom_transformations = {"square": lambda x: x**2}

    Returns:
        A table where all extracted attributes are merged based on a common index or None
        if no attributes were extracted. If the requested data_type is not present in
        graph_attrs, or if the attribute dict is empty, returns None. This is intended
        to allow optional annotation blocks.

    """

    if data_type not in ENTITIES_W_DATA:
        raise ValueError(
            f'"data_type" was {data_type} and must be in {", ".join(ENTITIES_W_DATA)}'
        )

    if data_type not in graph_attrs.keys():
        logger.info(
            f'No {data_type} annotations provided in "graph_attrs"; returning None'
        )
        return None

    entity_attrs = graph_attrs[data_type]
    # validating dict
    _validate_entity_attrs(entity_attrs, custom_transformations=custom_transformations)

    if len(entity_attrs) == 0:
        logger.info(
            f'No attributes defined for "{data_type}" in graph_attrs; returning None'
        )
        return None

    data_type_attr = data_type + "_data"
    entity_data_tbls = getattr(sbml_dfs, data_type_attr)

    data_list = list()
    for k, v in entity_attrs.items():
        # v["table"] is always present if entity_attrs is non-empty and validated
        if v["table"] not in entity_data_tbls.keys():
            raise ValueError(
                f"{v['table']} was defined as a table in \"graph_attrs\" but "
                f'it is not present in the "{data_type_attr}" of the sbml_dfs'
            )

        if v["variable"] not in entity_data_tbls[v["table"]].columns.tolist():
            raise ValueError(
                f"{v['variable']} was defined as a variable in \"graph_attrs\" but "
                f"it is not present in the {v['table']} of the \"{data_type_attr}\" of "
                "the sbml_dfs"
            )

        entity_series = entity_data_tbls[v["table"]][v["variable"]].rename(k)
        trans_name = v.get("trans", DEFAULT_WT_TRANS)
        # Look up transformation
        if custom_transformations and trans_name in custom_transformations:
            trans_fxn = custom_transformations[trans_name]
        elif trans_name in DEFINED_WEIGHT_TRANSFORMATION:
            trans_fxn = globals()[DEFINED_WEIGHT_TRANSFORMATION[trans_name]]
        else:
            # This should never be hit if _validate_entity_attrs is called correctly.
            raise ValueError(
                f"Transformation '{trans_name}' not found in custom_transformations or DEFINED_WEIGHT_TRANSFORMATION."
            )
        entity_series = entity_series.apply(trans_fxn)
        data_list.append(entity_series)

    if len(data_list) == 0:
        return None

    return pd.concat(data_list, axis=1)


def apply_weight_transformations(
    edges_df: pd.DataFrame, reaction_attrs: dict, custom_transformations: dict = None
):
    """
    Apply Weight Transformations

    Args:
        edges_df (pd.DataFrame): a table of edges and their attributes extracted
            from a cpr_grpah.
        reaction_attrs (dict):
            A dictionary of attributes identifying weighting attributes within
            an sbml_df's reaction_data, how they will be named in edges_df (the keys),
            and how they should be transformed (the "trans" aliases")
        custom_transformations (dict, optional):
            A dictionary mapping transformation names to functions. If provided, these
            will be checked before built-in transformations.

    Returns:
        transformed_edges_df (pd.DataFrame): edges_df with weight variables transformed.

    """

    _validate_entity_attrs(
        reaction_attrs, custom_transformations=custom_transformations
    )

    transformed_edges_df = copy.deepcopy(edges_df)
    for k, v in reaction_attrs.items():
        if k not in transformed_edges_df.columns:
            raise ValueError(f"A weighting variable {k} was missing from edges_df")

        trans_name = v["trans"]
        # Look up transformation
        if custom_transformations and trans_name in custom_transformations:
            trans_fxn = custom_transformations[trans_name]
        elif trans_name in DEFINED_WEIGHT_TRANSFORMATION:
            trans_fxn = globals()[DEFINED_WEIGHT_TRANSFORMATION[trans_name]]
        else:
            # This should never be hit if _validate_entity_attrs is called correctly.
            raise ValueError(
                f"Transformation '{trans_name}' not found in custom_transformations or DEFINED_WEIGHT_TRANSFORMATION."
            )

        transformed_edges_df[k] = transformed_edges_df[k].apply(trans_fxn)

    return transformed_edges_df


def summarize_weight_calibration(cpr_graph: ig.Graph, reaction_attrs: dict) -> None:
    """
    Summarize Weight Calibration

    For a network with multiple sources for edge weights summarize the alignment of
    different weighting schemes and how they map onto our notion of "good" versus
    "dubious" weights.

    Args:
        cpr_graph (ig.Graph): A graph where edge weights have already been calibrated.
        reaction_attrs (dict): a dictionary summarizing the types of weights that
            exist and how they are transformed for calibration.

    Returns:
        None

    """

    score_calibration_df = pd.DataFrame(SCORE_CALIBRATION_POINTS_DICT)
    score_calibration_df_calibrated = apply_weight_transformations(
        score_calibration_df, reaction_attrs
    )

    calibrated_edges = cpr_graph.get_edge_dataframe()

    _summarize_weight_calibration_table(
        calibrated_edges, score_calibration_df, score_calibration_df_calibrated
    )

    _summarize_weight_calibration_plots(
        calibrated_edges, score_calibration_df_calibrated
    )

    return None


def add_graph_weights(
    cpr_graph: ig.Graph,
    reaction_attrs: dict,
    weighting_strategy: str = CPR_WEIGHTING_STRATEGIES.UNWEIGHTED,
) -> ig.Graph:
    """
    Add Graph Weights

    Apply a weighting strategy to generate edge weights on a graph. For directed graphs "upstream_weights" will
    be generated as well which should be used when searching for a node's ancestors.

    Args:
        cpr_graph (ig.Graph): a graphical network of molecules/reactions (nodes) and edges linking them.
        reaction_attrs (dict): an optional dict
        weighting_strategy: a network weighting strategy with options:
            - unweighted: all weights (and upstream_weights for directed graphs) are set to 1.
            - topology: weight edges by the degree of the source nodes favoring nodes emerging from nodes
                with few connections.
            - mixed: transform edges with a quantitative score based on reaction_attrs; and set edges
                without quantitative score as a source-specific weight.
            - calibrated: transforme edges with a quantitative score based on reaction_attrs and combine them
                with topology scores to generate a consensus.

    """

    cpr_graph_updated = copy.deepcopy(cpr_graph)

    _validate_entity_attrs(reaction_attrs)

    if weighting_strategy not in VALID_WEIGHTING_STRATEGIES:
        raise ValueError(
            f"weighting_strategy was {weighting_strategy} and must be one of: "
            f"{', '.join(VALID_WEIGHTING_STRATEGIES)}"
        )

    # count parents and children and create weights based on them
    topology_weighted_graph = _create_topology_weights(cpr_graph_updated)

    if weighting_strategy == CPR_WEIGHTING_STRATEGIES.TOPOLOGY:
        topology_weighted_graph.es[CPR_GRAPH_EDGES.WEIGHTS] = (
            topology_weighted_graph.es["topo_weights"]
        )
        if cpr_graph_updated.is_directed():
            topology_weighted_graph.es[CPR_GRAPH_EDGES.UPSTREAM_WEIGHTS] = (
                topology_weighted_graph.es["upstream_topo_weights"]
            )

        return topology_weighted_graph

    if weighting_strategy == CPR_WEIGHTING_STRATEGIES.UNWEIGHTED:
        # set weights as a constant
        topology_weighted_graph.es[CPR_GRAPH_EDGES.WEIGHTS] = 1
        if cpr_graph_updated.is_directed():
            topology_weighted_graph.es[CPR_GRAPH_EDGES.UPSTREAM_WEIGHTS] = 1
        return topology_weighted_graph

    if weighting_strategy == CPR_WEIGHTING_STRATEGIES.MIXED:
        return _add_graph_weights_mixed(topology_weighted_graph, reaction_attrs)

    if weighting_strategy == CPR_WEIGHTING_STRATEGIES.CALIBRATED:
        return _add_graph_weights_calibration(topology_weighted_graph, reaction_attrs)

    raise ValueError(f"No logic implemented for {weighting_strategy}")


def _create_cpr_graph_bipartite(sbml_dfs: sbml_dfs_core.SBML_dfs) -> pd.DataFrame:
    """Turn an sbml_dfs model into a bipartite graph linking molecules to reactions."""

    # setup edges
    network_edges = (
        sbml_dfs.reaction_species.reset_index()[
            [SBML_DFS.R_ID, SBML_DFS.SC_ID, SBML_DFS.STOICHIOMETRY, SBML_DFS.SBO_TERM]
        ]
        # rename species and reactions to reflect from -> to edges
        .rename(
            columns={
                SBML_DFS.SC_ID: CPR_GRAPH_NODE_TYPES.SPECIES,
                SBML_DFS.R_ID: CPR_GRAPH_NODE_TYPES.REACTION,
            }
        )
    )
    # add back an r_id variable so that each edge is annotated by a reaction
    network_edges[CPR_GRAPH_EDGES.R_ID] = network_edges[CPR_GRAPH_NODE_TYPES.REACTION]

    # add edge weights
    cspecies_features = sbml_dfs.get_cspecies_features()
    network_edges = network_edges.merge(
        cspecies_features, left_on=CPR_GRAPH_NODE_TYPES.SPECIES, right_index=True
    )

    # if directed then flip substrates and modifiers to the origin edge
    edge_vars = network_edges.columns.tolist()

    origins = network_edges[network_edges[SBML_DFS.STOICHIOMETRY] <= 0]
    origin_edges = origins.loc[:, [edge_vars[1], edge_vars[0]] + edge_vars[2:]].rename(
        columns={
            CPR_GRAPH_NODE_TYPES.SPECIES: CPR_GRAPH_EDGES.FROM,
            CPR_GRAPH_NODE_TYPES.REACTION: CPR_GRAPH_EDGES.TO,
        }
    )

    dests = network_edges[network_edges[SBML_DFS.STOICHIOMETRY] > 0]
    dest_edges = dests.rename(
        columns={
            CPR_GRAPH_NODE_TYPES.REACTION: CPR_GRAPH_EDGES.FROM,
            CPR_GRAPH_NODE_TYPES.SPECIES: CPR_GRAPH_EDGES.TO,
        }
    )

    network_edges = pd.concat([origin_edges, dest_edges])

    return network_edges


def _create_cpr_graph_tiered(
    sbml_dfs: sbml_dfs_core.SBML_dfs, graph_type: str
) -> pd.DataFrame:
    """Turn an sbml_dfs model into a tiered graph which links upstream entities to downstream ones."""

    # check whether all expect SBO terms are present
    invalid_sbo_terms = sbml_dfs.reaction_species[
        ~sbml_dfs.reaction_species[SBML_DFS.SBO_TERM].isin(MINI_SBO_TO_NAME.keys())
    ]

    if invalid_sbo_terms.shape[0] != 0:
        invalid_counts = invalid_sbo_terms.value_counts(SBML_DFS.SBO_TERM).to_frame("N")
        if not isinstance(invalid_counts, pd.DataFrame):
            raise TypeError("invalid_counts must be a pandas DataFrame")
        logger.warning(utils.style_df(invalid_counts, headers="keys"))  # type: ignore
        raise ValueError("Some reaction species have unusable SBO terms")

    # load and validate the schema of graph_type
    graph_hierarchy_df = _create_graph_hierarchy_df(graph_type)

    # organize reaction species for defining connections
    sorted_reaction_species = sbml_dfs.reaction_species.set_index(
        [SBML_DFS.R_ID, SBML_DFS.SBO_TERM]
    ).sort_index()

    logger.info(
        f"Formatting {sorted_reaction_species.shape[0]} reactions species as "
        "tiered edges."
    )

    # infer tiered edges in each reaction
    all_reaction_edges = [
        _format_tiered_reaction_species(
            r, sorted_reaction_species, sbml_dfs, graph_hierarchy_df
        )
        for r in sorted_reaction_species.index.get_level_values(SBML_DFS.R_ID).unique()
    ]
    all_reaction_edges_df = pd.concat(all_reaction_edges).reset_index(drop=True)

    # test for reactions missing substrates
    r_id_list = sorted_reaction_species.index.get_level_values(0).unique()
    r_id_reactant_only = [
        x for x in r_id_list if len(sorted_reaction_species.loc[x]) == 1
    ]

    if len(r_id_reactant_only) > 0:
        logger.warning(f"{len(r_id_reactant_only)} reactions are missing substrates")
        all_reaction_edges_df_pre = all_reaction_edges_df.copy()
        all_reaction_edges_df = all_reaction_edges_df_pre[
            ~all_reaction_edges_df_pre[SBML_DFS.R_ID].isin(r_id_reactant_only)
        ]

    logger.info(
        "Adding additional attributes to edges, e.g., # of children and parents."
    )

    # add compartmentalized species summaries to weight edges
    cspecies_features = sbml_dfs.get_cspecies_features()

    # calculate undirected and directed degrees (i.e., # of parents and children)
    # based on a network's edgelist. this used when the network representation is
    # not the bipartite network which can be trivially obtained from the pathway
    # specification
    unique_edges = (
        all_reaction_edges_df.groupby([CPR_GRAPH_EDGES.FROM, CPR_GRAPH_EDGES.TO])
        .first()
        .reset_index()
    )

    # children
    n_children = (
        unique_edges[CPR_GRAPH_EDGES.FROM]
        .value_counts()
        # rename values to the child name
        .to_frame(name=CPR_GRAPH_EDGES.SC_CHILDREN)
        .reset_index()
        .rename(
            {
                CPR_GRAPH_EDGES.FROM: SBML_DFS.SC_ID,
            },
            axis=1,
        )
    )

    # parents
    n_parents = (
        unique_edges[CPR_GRAPH_EDGES.TO]
        .value_counts()
        # rename values to the parent name
        .to_frame(name=CPR_GRAPH_EDGES.SC_PARENTS)
        .reset_index()
        .rename(
            {
                CPR_GRAPH_EDGES.TO: SBML_DFS.SC_ID,
            },
            axis=1,
        )
    )

    graph_degree_by_edgelist = n_children.merge(n_parents, how="outer").fillna(int(0))

    graph_degree_by_edgelist[CPR_GRAPH_EDGES.SC_DEGREE] = (
        graph_degree_by_edgelist[CPR_GRAPH_EDGES.SC_CHILDREN]
        + graph_degree_by_edgelist[CPR_GRAPH_EDGES.SC_PARENTS]
    )
    graph_degree_by_edgelist = (
        graph_degree_by_edgelist[
            ~graph_degree_by_edgelist[SBML_DFS.SC_ID].str.contains("R[0-9]{8}")
        ]
        .set_index(SBML_DFS.SC_ID)
        .sort_index()
    )

    cspecies_features = (
        cspecies_features.drop(
            [
                CPR_GRAPH_EDGES.SC_DEGREE,
                CPR_GRAPH_EDGES.SC_CHILDREN,
                CPR_GRAPH_EDGES.SC_PARENTS,
            ],
            axis=1,
        )
        .join(graph_degree_by_edgelist)
        .fillna(int(0))
    )

    is_from_reaction = all_reaction_edges_df[CPR_GRAPH_EDGES.FROM].isin(
        sbml_dfs.reactions.index.tolist()
    )
    is_from_reaction = all_reaction_edges_df[CPR_GRAPH_EDGES.FROM].isin(
        sbml_dfs.reactions.index
    )
    # add substrate weight whenever "from" edge is a molecule
    # and product weight when the "from" edge is a reaction
    decorated_all_reaction_edges_df = pd.concat(
        [
            all_reaction_edges_df[~is_from_reaction].merge(
                cspecies_features, left_on=CPR_GRAPH_EDGES.FROM, right_index=True
            ),
            all_reaction_edges_df[is_from_reaction].merge(
                cspecies_features, left_on=CPR_GRAPH_EDGES.TO, right_index=True
            ),
        ]
    ).sort_index()

    if all_reaction_edges_df.shape[0] != decorated_all_reaction_edges_df.shape[0]:
        msg = (
            "'decorated_all_reaction_edges_df' and 'all_reaction_edges_df' should\n"
            "have the same number of rows but they did not"
        )

        raise ValueError(msg)

    logger.info(f"Done preparing {graph_type} graph")

    return decorated_all_reaction_edges_df


def _format_tiered_reaction_species(
    r_id: str,
    sorted_reaction_species: pd.DataFrame,
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    graph_hierarchy_df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Format Tiered Reaction Species

    Refactor a reaction's species into tiered edges between substrates, products, enzymes and allosteric regulators.
    """

    rxn_species = sorted_reaction_species.loc[r_id]
    if not isinstance(rxn_species, pd.DataFrame):
        raise TypeError("rxn_species must be a pandas DataFrame")
    if list(rxn_species.index.names) != [SBML_DFS.SBO_TERM]:
        raise ValueError("rxn_species index names must be [SBML_DFS.SBO_TERM]")
    if rxn_species.columns.tolist() != [SBML_DFS.SC_ID, SBML_DFS.STOICHIOMETRY]:
        raise ValueError(
            "rxn_species columns must be [SBML_DFS.SC_ID, SBML_DFS.STOICHIOMETRY]"
        )

    rxn_sbo_terms = set(rxn_species.index.unique())
    # map to common names
    rxn_sbo_names = {MINI_SBO_TO_NAME[x] for x in rxn_sbo_terms}

    # is the reaction a general purpose interaction
    if len(rxn_sbo_names) == 1:
        if list(rxn_sbo_names)[0] == SBOTERM_NAMES.INTERACTOR:
            # further validation happens in the function - e.g., exactly two interactors
            return _format_interactors_for_tiered_graph(r_id, rxn_species, sbml_dfs)

    if SBOTERM_NAMES.INTERACTOR in rxn_sbo_names:
        logger.warning(
            f"Invalid combinations of SBO_terms in {str(r_id)} : {sbml_dfs.reactions.loc[r_id][SBML_DFS.R_NAME]}. "
            "If interactors are present then there can't be any other types of reaction species. "
            f"The following roles were defined: {', '.join(rxn_sbo_names)}"
        )

    # reorganize molecules and the reaction itself into tiers
    entities_ordered_by_tier = (
        pd.concat(
            [
                (
                    rxn_species.reset_index()
                    .rename({SBML_DFS.SC_ID: "entity_id"}, axis=1)
                    .merge(graph_hierarchy_df)
                ),
                graph_hierarchy_df[
                    graph_hierarchy_df[CPR_GRAPH_EDGES.SBO_NAME]
                    == CPR_GRAPH_NODE_TYPES.REACTION
                ].assign(entity_id=r_id, r_id=r_id),
            ]
        )
        .sort_values(["tier"])
        .set_index("tier")
    )
    ordered_tiers = entities_ordered_by_tier.index.get_level_values("tier").unique()

    if len(ordered_tiers) <= 1:
        raise ValueError("ordered_tiers must have more than one element")

    # which tier is the reaction?
    reaction_tier = graph_hierarchy_df["tier"][
        graph_hierarchy_df[CPR_GRAPH_EDGES.SBO_NAME] == CPR_GRAPH_NODE_TYPES.REACTION
    ].tolist()[0]

    rxn_edges = list()
    past_reaction = False
    for i in range(0, len(ordered_tiers) - 1):
        formatted_tier_combo = _format_tier_combo(
            entities_ordered_by_tier.loc[[ordered_tiers[i]]],
            entities_ordered_by_tier.loc[[ordered_tiers[i + 1]]],
            past_reaction,
        )

        if ordered_tiers[i + 1] == reaction_tier:
            past_reaction = True

        rxn_edges.append(formatted_tier_combo)

    rxn_edges_df = (
        pd.concat(rxn_edges)[
            [
                CPR_GRAPH_EDGES.FROM,
                CPR_GRAPH_EDGES.TO,
                CPR_GRAPH_EDGES.STOICHIOMETRY,
                CPR_GRAPH_EDGES.SBO_TERM,
            ]
        ]
        .reset_index(drop=True)
        .assign(r_id=r_id)
    )

    return rxn_edges_df


def _format_tier_combo(
    upstream_tier: pd.DataFrame, downstream_tier: pd.DataFrame, past_reaction: bool
) -> pd.DataFrame:
    """
    Format Tier Combo

    Create a set of edges crossing two tiers of a tiered graph. This will involve an
      all x all combination of entries. Tiers form an ordering along the molecular entities
      in a reaction plus a tier for the reaction itself. Attributes such as stoichiometry
      and sbo_term will be passed from the tier which is furthest from the reaction tier
      to ensure that each tier of molecular data applies its attributes to a single set of
      edges while the "reaction" tier does not. Reaction entities have neither a
      stoichiometery or sbo_term annotation.

    Args:
        upstream_tier (pd.DataFrame): A table containing upstream entities in a reaction,
            e.g., regulators.
        downstream_tier (pd.DataFrame): A table containing downstream entities in a reaction,
            e.g., catalysts.
        past_reaction (bool): if True then attributes will be taken from downstream_tier and
            if False they will come from upstream_tier.

    Returns:
        formatted_tier_combo (pd.DataFrame): A table of edges containing (from, to, stoichiometry, sbo_term, r_id). The
        number of edges is the product of the number of entities in the upstream tier
        times the number in the downstream tier.

    """

    upstream_fields = ["entity_id", SBML_DFS.STOICHIOMETRY, SBML_DFS.SBO_TERM]
    downstream_fields = ["entity_id"]

    if past_reaction:
        # swap fields
        upstream_fields, downstream_fields = downstream_fields, upstream_fields

    formatted_tier_combo = (
        upstream_tier[upstream_fields]
        .rename({"entity_id": CPR_GRAPH_EDGES.FROM}, axis=1)
        .assign(_joiner=1)
    ).merge(
        (
            downstream_tier[downstream_fields]
            .rename({"entity_id": CPR_GRAPH_EDGES.TO}, axis=1)
            .assign(_joiner=1)
        ),
        left_on="_joiner",
        right_on="_joiner",
    )

    return formatted_tier_combo


def _create_graph_hierarchy_df(graph_type: str) -> pd.DataFrame:
    """
    Create Graph Hierarchy DataFrame

    Format a graph hierarchy list of lists and a pd.DataFrame

    Args:
        graph_type (str):
            The type of tiered graph to work with. Each type has its own specification in constants.py.

    Returns:
        A pandas DataFrame with sbo_name, tier, and sbo_term.

    """

    if graph_type == CPR_GRAPH_TYPES.REGULATORY:
        sbo_names_hierarchy = REGULATORY_GRAPH_HIERARCHY
    elif graph_type == CPR_GRAPH_TYPES.SURROGATE:
        sbo_names_hierarchy = SURROGATE_GRAPH_HIERARCHY
    else:
        raise NotImplementedError(f"{graph_type} is not a valid graph_type")

    # format as a DF
    graph_hierarchy_df = pd.concat(
        [
            pd.DataFrame({"sbo_name": sbo_names_hierarchy[i]}).assign(tier=i)
            for i in range(0, len(sbo_names_hierarchy))
        ]
    ).reset_index(drop=True)
    graph_hierarchy_df[SBML_DFS.SBO_TERM] = graph_hierarchy_df["sbo_name"].apply(
        lambda x: MINI_SBO_FROM_NAME[x] if x != CPR_GRAPH_NODE_TYPES.REACTION else ""
    )

    # ensure that the output is expected
    utils.match_pd_vars(
        graph_hierarchy_df,
        req_vars={CPR_GRAPH_EDGES.SBO_NAME, "tier", SBML_DFS.SBO_TERM},
        allow_series=False,
    ).assert_present()

    return graph_hierarchy_df


def _add_graph_weights_mixed(cpr_graph: ig.Graph, reaction_attrs: dict) -> ig.Graph:
    """Weight a graph using a mixed approach combining source-specific weights and existing edge weights."""

    edges_df = cpr_graph.get_edge_dataframe()

    calibrated_edges = apply_weight_transformations(edges_df, reaction_attrs)
    calibrated_edges = _create_source_weights(calibrated_edges, "source_wt")

    score_vars = list(reaction_attrs.keys())
    score_vars.append("source_wt")

    logger.info(f"Creating mixed scores based on {', '.join(score_vars)}")

    calibrated_edges["weights"] = calibrated_edges[score_vars].min(axis=1)

    cpr_graph.es[CPR_GRAPH_EDGES.WEIGHTS] = calibrated_edges[CPR_GRAPH_EDGES.WEIGHTS]
    if cpr_graph.is_directed():
        cpr_graph.es[CPR_GRAPH_EDGES.UPSTREAM_WEIGHTS] = calibrated_edges[
            CPR_GRAPH_EDGES.WEIGHTS
        ]

    # add other attributes and update transformed attributes
    cpr_graph.es["source_wt"] = calibrated_edges["source_wt"]
    for k in reaction_attrs.keys():
        cpr_graph.es[k] = calibrated_edges[k]

    return cpr_graph


def _add_graph_weights_calibration(
    cpr_graph: ig.Graph, reaction_attrs: dict
) -> ig.Graph:
    """Weight a graph using a calibrated strategy which aims to roughly align qualiatively similar weights from different sources."""

    edges_df = cpr_graph.get_edge_dataframe()

    calibrated_edges = apply_weight_transformations(edges_df, reaction_attrs)

    score_vars = list(reaction_attrs.keys())
    score_vars.append("topo_weights")

    logger.info(f"Creating calibrated scores based on {', '.join(score_vars)}")
    cpr_graph.es["weights"] = calibrated_edges[score_vars].min(axis=1)

    if cpr_graph.is_directed():
        score_vars = list(reaction_attrs.keys())
        score_vars.append("upstream_topo_weights")
        cpr_graph.es["upstream_weights"] = calibrated_edges[score_vars].min(axis=1)

    # add other attributes and update transformed attributes
    for k in reaction_attrs.keys():
        cpr_graph.es[k] = calibrated_edges[k]

    return cpr_graph


def _add_edge_attr_to_vertex_graph(
    cpr_graph: ig.Graph,
    edge_attr_list: list,
    shared_node_key: str = "r_id",
) -> ig.Graph:
    """
    Merge edge attribute(s) from edge_attr_list to vetices of an igraph

    Parameters
    ----------
    cpr_graph : iGraph
        A graph generated by create_cpr_graph()
    edge_attr_list: list
        A list containing attributes to pull out of edges, then to add to vertices
    shared_node_key : str
        key in edge that is shared with vertex, to map edge ids to corresponding vertex ids

    Returns:
    ----------
    An Igraph network
    """

    if len(edge_attr_list) == 0:
        logger.warning(
            "No edge attributes were passed, " "thus return the input graph."
        )
        return cpr_graph

    graph_vertex_df = cpr_graph.get_vertex_dataframe()
    graph_edge_df = cpr_graph.get_edge_dataframe()

    if shared_node_key not in graph_edge_df.columns.to_list():
        logger.warning(
            f"{shared_node_key} is not in the current edge attributes. "
            "shared_node_key must be an existing edge attribute"
        )
        return cpr_graph

    graph_edge_df_sub = graph_edge_df.loc[:, [shared_node_key] + edge_attr_list].copy()

    # check whether duplicated edge ids by shared_node_key have the same attribute values.
    # If not, give warning, and keep the first value. (which can be improved later)
    check_edgeid_attr_unique = (
        graph_edge_df_sub.groupby(shared_node_key)[edge_attr_list].nunique() == 1
    )

    # check any False in check_edgeid_attr_unique's columns, if so, get the column names
    bool_edgeid_attr_unique = (check_edgeid_attr_unique.isin([False])).any()  # type: ignore

    non_unique_indices = [
        i for i, value in enumerate(bool_edgeid_attr_unique.to_list()) if value
    ]

    # if edge ids with duplicated shared_node_key have more than 1 unique values
    # for attributes of interest
    non_unique_egde_attr = bool_edgeid_attr_unique.index[non_unique_indices].to_list()

    if len(non_unique_egde_attr) == 0:
        logger.info("Per duplicated edge ids, attributes have only 1 unique value.")
    else:
        logger.warning(
            f"Per duplicated edge ids, attributes: {non_unique_egde_attr} "
            "contain more than 1 unique values"
        )

    # remove duplicated edge attribute values
    graph_edge_df_sub_no_duplicate = graph_edge_df_sub.drop_duplicates(
        subset=shared_node_key, keep="first"
    )

    # rename shared_node_key to vertex key 'name'
    # as in net_create.create_cpr_graph(), vertex_name_attr is set to 'name'
    graph_edge_df_sub_no_duplicate = graph_edge_df_sub_no_duplicate.rename(
        columns={shared_node_key: "name"},
    )

    # merge edge attributes in graph_edge_df_sub_no_duplicate to vertex_df,
    # by shared key 'name'
    graph_vertex_df_w_edge_attr = pd.merge(
        graph_vertex_df,
        graph_edge_df_sub_no_duplicate,
        on="name",
        how="outer",
    )

    logger.info(f"Adding {edge_attr_list} to vertex attributes")
    # Warning for NaN values in vertex attributes:
    if graph_vertex_df_w_edge_attr.isnull().values.any():
        logger.warning(
            "NaN values are present in the newly added vertex attributes. "
            "Please assign proper values to those vertex attributes."
        )

    # assign the edge_attrs from edge_attr_list to cpr_graph's vertices:
    # keep the same edge attribute names:
    for col_name in edge_attr_list:
        cpr_graph.vs[col_name] = graph_vertex_df_w_edge_attr[col_name]

    return cpr_graph


def _summarize_weight_calibration_table(
    calibrated_edges: pd.DataFrame,
    score_calibration_df: pd.DataFrame,
    score_calibration_df_calibrated: pd.DataFrame,
):
    """Create a table comparing edge weights from multiple sources."""

    # generate a table summarizing different scoring measures
    #
    # a set of calibration points defined in DEFINED_WEIGHT_TRANSFORMATION which map
    # onto what we might consider strong versus dubious edges are compared to the
    # observed scores to see whether these calibration points generally map onto
    # the expected quantiles of the score distribution.
    #
    # different scores are also compared to see whether there calibrations are generally
    # aligned. that is to say a strong weight based on one scoring measure would receive
    # a similar quantitative score to a strong score for another measure.

    score_calibration_long_raw = (
        score_calibration_df.reset_index()
        .rename({"index": "edge_strength"}, axis=1)
        .melt(
            id_vars="edge_strength", var_name="weight_measure", value_name="raw_weight"
        )
    )

    score_calibration_long_calibrated = (
        score_calibration_df_calibrated.reset_index()
        .rename({"index": "edge_strength"}, axis=1)
        .melt(
            id_vars="edge_strength",
            var_name="weight_measure",
            value_name="trans_weight",
        )
    )

    score_calibration_table_long = score_calibration_long_raw.merge(
        score_calibration_long_calibrated
    )

    # compare calibration points to the quantiles of the observed score distributions
    score_quantiles = list()
    for ind, row in score_calibration_table_long.iterrows():
        score_quantiles.append(
            1
            - np.mean(
                calibrated_edges[row["weight_measure"]].dropna() >= row["trans_weight"]
            )
        )
    score_calibration_table_long["quantile_of_score_dist"] = score_quantiles

    return utils.style_df(score_calibration_table_long, headers="keys")


def _summarize_weight_calibration_plots(
    calibrated_edges: pd.DataFrame, score_calibration_df_calibrated: pd.DataFrame
) -> None:
    """Create a couple of plots summarizing the relationships between different scoring measures."""

    # set up a 2 x 1 plot
    f, (ax1, ax2) = plt.subplots(1, 2)

    calibrated_edges[["topo_weights", "string_wt"]].plot(
        kind="hist", bins=50, alpha=0.5, ax=ax1
    )
    ax1.set_title("Distribution of scores\npost calibration")

    score_calibration_df_calibrated.plot("weights", "string_wt", kind="scatter", ax=ax2)

    for k, v in score_calibration_df_calibrated.iterrows():
        ax2.annotate(k, v)
    ax2.axline((0, 0), slope=1.0, color="C0", label="by slope")
    ax2.set_title("Comparing STRING and\nTopology calibration points")

    return None


def _create_source_weights(
    edges_df: pd.DataFrame,
    source_wt_var: str = "source_wt",
    source_vars_dict: dict = SOURCE_VARS_DICT,
    source_wt_default: int = 1,
) -> pd.DataFrame:
    """ "
    Create Source Weights

    Create weights based on an edges source. This is a simple but crude way of allowing different
    data sources to have different support if we think that some are more trustworthly than others.

    Args:
        edges_df: pd.DataFrame
            The edges dataframe to add the source weights to.
        source_wt_var: str
            The name of the column to store the source weights.
        source_vars_dict: dict
            Dictionary with keys indicating edge attributes and values indicating the weight to assign
            to that attribute. This value is generally the largest weight that can be assigned to an
            edge so that the numeric weight is chosen over the default.
        source_wt_default: int
            The default weight to assign to an edge if no other weight attribute is found.

    Returns:
        pd.DataFrame
            The edges dataframe with the source weights added.

    """

    logger.warning(
        "_create_source_weights should be reimplemented once https://github.com/calico/pathadex-data/issues/95 "
        "is fixed. The current implementation is quite limited."
    )

    # currently, we will look for values of source_indicator_var which are non NA and set them to
    # source_indicator_match_score and setting entries which are NA as source_indicator_nonmatch_score.
    #
    # this is a simple way of flagging string vs. non-string scores

    included_weight_vars = set(source_vars_dict.keys()).intersection(
        set(edges_df.columns)
    )
    if len(included_weight_vars) == 0:
        logger.warning(
            f"No edge attributes were found which match those in source_vars_dict: {', '.join(source_vars_dict.keys())}"
        )
        edges_df[source_wt_var] = source_wt_default
        return edges_df

    edges_df_source_wts = edges_df[list(included_weight_vars)].copy()
    for wt in list(included_weight_vars):
        edges_df_source_wts[wt] = [
            source_wt_default if x is True else source_vars_dict[wt]
            for x in edges_df[wt].isna()
        ]

    source_wt_edges_df = edges_df.join(
        edges_df_source_wts.max(axis=1).rename(source_wt_var)
    )

    return source_wt_edges_df


def _wt_transformation_identity(x):
    """Identity"""
    return x


def _wt_transformation_string(x):
    """Map STRING scores to a similar scale as topology weights."""

    return 250000 / np.power(x, 1.7)


def _wt_transformation_string_inv(x):
    """Map STRING scores so they work with source weights."""

    # string scores are bounded on [0, 1000]
    # and score/1000 is roughly a probability that
    # there is a real interaction (physical, genetic, ...)
    # reported string scores are currently on [150, 1000]
    # so this transformation will map these onto {6.67, 1}

    return 1 / (x / 1000)


def _format_interactors_for_tiered_graph(
    r_id: str, rxn_species: pd.DataFrame, sbml_dfs: sbml_dfs_core.SBML_dfs
) -> pd.DataFrame:
    """Format an undirected interactions for tiered graph so interactions are linked even though they would be on the same tier."""

    interactor_data = rxn_species.loc[MINI_SBO_FROM_NAME["interactor"]]
    if interactor_data.shape[0] != 2:
        raise ValueError(
            f"{interactor_data.shape[0]} interactors present for {str(r_id)} : "
            f"{sbml_dfs.reactions.loc[r_id]['r_name']}. "
            "Reactions with interactors must have exactly two interactors"
        )

    if not (interactor_data["stoichiometry"] == 0).any():
        raise ValueError(
            f"Interactors had non-zero stoichiometry for {str(r_id)} : {sbml_dfs.reactions.loc[r_id]['r_name']}. "
            "If stoichiometry is important for this reaction then it should use other SBO terms "
            "(e.g., substrate and product)."
        )

    # set the first entry as "from" and second as "to" if stoi is zero.
    # the reverse reaction will generally be added later because these
    # reactions should be reversible

    return pd.DataFrame(
        {
            "from": interactor_data["sc_id"].iloc[0],
            "to": interactor_data["sc_id"].iloc[1],
            "sbo_term": MINI_SBO_FROM_NAME["interactor"],
            "stoichiometry": 0,
            "r_id": r_id,
        },
        index=[0],
    )


def _add_graph_species_attribute(
    cpr_graph: ig.Graph,
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    species_graph_attrs: dict,
    custom_transformations: Optional[dict] = None,
) -> ig.Graph:
    """
    Add meta-data from species_data to existing igraph's vertices.

    This function augments the vertices of an igraph network with additional attributes
    derived from the species-level data in the provided SBML_dfs object. The attributes
    to add are specified in the species_graph_attrs dictionary, and can be transformed
    using either built-in or user-supplied transformation functions.

    Parameters
    ----------
    cpr_graph : ig.Graph
        The igraph network to augment.
    sbml_dfs : sbml_dfs_core.SBML_dfs
        The SBML_dfs object containing species data.
    species_graph_attrs : dict
        Dictionary specifying which attributes to pull from species_data and how to transform them.
        The structure should be {attribute_name: {"table": ..., "variable": ..., "trans": ...}}.
    custom_transformations : dict, optional
        Dictionary mapping transformation names to functions. If provided, these will be checked
        before built-in transformations. Example: {"square": lambda x: x**2}

    Returns
    -------
    ig.Graph
        The input igraph network with additional vertex attributes added from species_data.
    """
    if not isinstance(species_graph_attrs, dict):
        raise TypeError(
            f"species_graph_attrs must be a dict, but was {type(species_graph_attrs)}"
        )

    # fail fast if species_graph_attrs is not properly formatted
    # also flatten attribute list to be added to vertex nodes
    sp_graph_key_list = []
    sp_node_attr_list = []
    for k in species_graph_attrs.keys():
        _validate_entity_attrs(
            species_graph_attrs[k], custom_transformations=custom_transformations
        )

        sp_graph_key_list.append(k)
        sp_node_attr_list.append(list(species_graph_attrs[k].keys()))

    # flatten sp_node_attr_list
    flat_sp_node_attr_list = [item for items in sp_node_attr_list for item in items]

    logger.info("Adding meta-data from species_data")

    curr_network_nodes_df = cpr_graph.get_vertex_dataframe()

    # add species-level attributes to nodes dataframe
    augmented_network_nodes_df = _augment_network_nodes(
        curr_network_nodes_df,
        sbml_dfs,
        species_graph_attrs,
        custom_transformations=custom_transformations,
    )

    for vs_attr in flat_sp_node_attr_list:
        # in case more than one vs_attr in the flat_sp_node_attr_list
        logger.info(f"Adding new attribute {vs_attr} to vertices")
        cpr_graph.vs[vs_attr] = augmented_network_nodes_df[vs_attr].values

    return cpr_graph


def _augment_network_nodes(
    network_nodes: pd.DataFrame,
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    species_graph_attrs: dict = dict(),
    custom_transformations: Optional[dict] = None,
) -> pd.DataFrame:
    """
    Add species-level attributes, expand network_nodes with s_id and c_id and then map to species-level attributes by s_id.

    This function merges species-level attributes from sbml_dfs into the provided network_nodes DataFrame,
    using the mapping in species_graph_attrs. Optionally, custom transformation functions can be provided
    to transform the attributes as they are added.

    Parameters
    ----------
    network_nodes : pd.DataFrame
        DataFrame of network nodes. Must include columns 'name', 'node_name', and 'node_type'.
    sbml_dfs : sbml_dfs_core.SBML_dfs
        The SBML_dfs object containing species data.
    species_graph_attrs : dict
        Dictionary specifying which attributes to pull from species_data and how to transform them.
        The structure should be {attribute_name: {"table": ..., "variable": ..., "trans": ...}}.
    custom_transformations : dict, optional
        Dictionary mapping transformation names to functions. If provided, these will be checked
        before built-in transformations. Example: {"square": lambda x: x**2}

    Returns
    -------
    pd.DataFrame
        The input network_nodes DataFrame with additional columns for each extracted and transformed attribute.
    """
    REQUIRED_NETWORK_NODE_ATTRS = {
        "name",
        "node_name",
        "node_type",
    }

    missing_required_network_nodes_attrs = REQUIRED_NETWORK_NODE_ATTRS.difference(
        set(network_nodes.columns.tolist())
    )
    if len(missing_required_network_nodes_attrs) > 0:
        raise ValueError(
            f"{len(missing_required_network_nodes_attrs)} required attributes were missing "
            "from network_nodes: "
            f"{', '.join(missing_required_network_nodes_attrs)}"
        )

    # include matching s_ids and c_ids of sc_ids
    # (the index of network_nodes df) in network_nodes df
    network_nodes_sid = pd.merge(
        network_nodes,
        sbml_dfs.compartmentalized_species[["s_id", "c_id"]],
        left_on="name",
        right_index=True,
        how="left",
    )

    # assign species_data related attributes to s_id
    species_graph_data = pluck_entity_data(
        sbml_dfs,
        species_graph_attrs,
        "species",
        custom_transformations=custom_transformations,
    )

    if species_graph_data is not None:
        # add species_graph_data to the network_nodes df, based on s_id
        network_nodes_wdata = network_nodes_sid.merge(
            species_graph_data, left_on="s_id", right_index=True, how="left"
        )
    else:
        network_nodes_wdata = network_nodes_sid

    # Note: multiple sc_ids with the same s_id will be assign with the same species_graph_data

    network_nodes_wdata = network_nodes_wdata.fillna(int(0)).drop(
        columns=["s_id", "c_id"]
    )

    return network_nodes_wdata


def _augment_network_edges(
    network_edges: pd.DataFrame,
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    reaction_graph_attrs: dict = dict(),
    custom_transformations: Optional[dict] = None,
) -> pd.DataFrame:
    """Add reversibility and other metadata from reactions.

    Parameters
    ----------
    network_edges : pd.DataFrame
        DataFrame of network edges.
    sbml_dfs : sbml_dfs_core.SBML_dfs
        The SBML_dfs object containing reaction data.
    reaction_graph_attrs : dict
        Dictionary of reaction attributes to add.
    custom_transformations : dict, optional
        Dictionary of custom transformation functions to use for attribute transformation.
    """
    REQUIRED_NETWORK_EDGE_ATTRS = {
        "from",
        "to",
        "stoichiometry",
        "sbo_term",
        "sc_degree",
        "sc_children",
        "sc_parents",
        "species_type",
        "r_id",
    }

    missing_required_network_edges_attrs = REQUIRED_NETWORK_EDGE_ATTRS.difference(
        set(network_edges.columns.tolist())
    )
    if len(missing_required_network_edges_attrs) > 0:
        raise ValueError(
            f"{len(missing_required_network_edges_attrs)} required attributes were missing "
            "from network_edges: "
            f"{', '.join(missing_required_network_edges_attrs)}"
        )

    network_edges = (
        network_edges[list(REQUIRED_NETWORK_EDGE_ATTRS)]
        # add reaction-level attributes
        .merge(
            sbml_dfs.reactions[SBML_DFS.R_ISREVERSIBLE],
            left_on=SBML_DFS.R_ID,
            right_index=True,
        )
    )

    # add other attributes based on reactions data
    reaction_graph_data = pluck_entity_data(
        sbml_dfs,
        reaction_graph_attrs,
        SBML_DFS.REACTIONS,
        custom_transformations=custom_transformations,
    )
    if reaction_graph_data is not None:
        network_edges = network_edges.merge(
            reaction_graph_data, left_on=SBML_DFS.R_ID, right_index=True, how="left"
        )

    return network_edges


def _reverse_network_edges(augmented_network_edges: pd.DataFrame) -> pd.DataFrame:
    """Flip reversible reactions to derive the reverse reaction."""

    # validate inputs
    missing_required_vars = CPR_GRAPH_REQUIRED_EDGE_VARS.difference(
        set(augmented_network_edges.columns.tolist())
    )

    if len(missing_required_vars) > 0:
        raise ValueError(
            "augmented_network_edges is missing required variables: "
            f"{', '.join(missing_required_vars)}"
        )

    # select all edges derived from reversible reactions
    reversible_reaction_edges = augmented_network_edges[
        augmented_network_edges[CPR_GRAPH_EDGES.R_ISREVERSIBLE]
    ]

    r_reaction_edges = (
        # ignore edges which start in a regulator or catalyst; even for a reversible reaction it
        # doesn't make sense for a regulator to be impacted by a target
        reversible_reaction_edges[
            ~reversible_reaction_edges[CPR_GRAPH_EDGES.SBO_TERM].isin(
                [
                    MINI_SBO_FROM_NAME[x]
                    for x in SBO_MODIFIER_NAMES.union({SBOTERM_NAMES.CATALYST})
                ]
            )
        ]
        # flip parent and child attributes
        .rename(
            {
                CPR_GRAPH_EDGES.FROM: CPR_GRAPH_EDGES.TO,
                CPR_GRAPH_EDGES.TO: CPR_GRAPH_EDGES.FROM,
                CPR_GRAPH_EDGES.SC_CHILDREN: CPR_GRAPH_EDGES.SC_PARENTS,
                CPR_GRAPH_EDGES.SC_PARENTS: CPR_GRAPH_EDGES.SC_CHILDREN,
            },
            axis=1,
        )
    )

    # switch substrates and products
    r_reaction_edges[CPR_GRAPH_EDGES.STOICHIOMETRY] = r_reaction_edges[
        CPR_GRAPH_EDGES.STOICHIOMETRY
    ].apply(
        # the ifelse statement prevents 0 being converted to -0 ...
        lambda x: -1 * x if x != 0 else 0
    )

    transformed_r_reaction_edges = pd.concat(
        [
            (
                r_reaction_edges[
                    r_reaction_edges[CPR_GRAPH_EDGES.SBO_TERM]
                    == MINI_SBO_FROM_NAME[SBOTERM_NAMES.REACTANT]
                ].assign(sbo_term=MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT])
            ),
            (
                r_reaction_edges[
                    r_reaction_edges[CPR_GRAPH_EDGES.SBO_TERM]
                    == MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT]
                ].assign(sbo_term=MINI_SBO_FROM_NAME[SBOTERM_NAMES.REACTANT])
            ),
            r_reaction_edges[
                ~r_reaction_edges[CPR_GRAPH_EDGES.SBO_TERM].isin(
                    [
                        MINI_SBO_FROM_NAME[SBOTERM_NAMES.REACTANT],
                        MINI_SBO_FROM_NAME[SBOTERM_NAMES.PRODUCT],
                    ]
                )
            ],
        ]
    )

    if transformed_r_reaction_edges.shape[0] != r_reaction_edges.shape[0]:
        raise ValueError(
            "transformed_r_reaction_edges and r_reaction_edges must have the same number of rows"
        )

    return transformed_r_reaction_edges.assign(
        direction=CPR_GRAPH_EDGE_DIRECTIONS.REVERSE
    )


def _create_topology_weights(
    cpr_graph: ig.Graph,
    base_score: float = 2,
    protein_multiplier: int = 1,
    metabolite_multiplier: int = 3,
    unknown_multiplier: int = 10,
    scale_multiplier_by_meandegree: bool = True,
) -> ig.Graph:
    """
    Create Topology Weights

    Add weights to a network based on its topology. Edges downstream of nodes
    with many connections receive a higher weight suggesting that any one
    of them is less likely to be regulatory. This is a simple and clearly
    flawed heuristic which can be combined with more principled weighting
    schemes.

    Args:
        cpr_graph (ig.Graph): a graph containing connections between molecules, proteins, and reactions.
        base_score (float): offset which will be added to all weights.
        protein_multiplier (int): multiplier for non-metabolite species (lower weight paths will tend to be selected).
        metabolite_multiplier (int): multiplier for metabolites [defined a species with a ChEBI ID).
        unknown_multiplier (int): multiplier for species without any identifier. See sbml_dfs_core.species_type_types.
        scale_multiplier_by_meandegree (bool): if True then multipliers will be rescaled by the average number of
            connections a node has (i.e., its degree) so that weights will be relatively similar regardless of network
            size and sparsity.

    Returns:
        cpr_graph (ig.Graph): graph with added topology weights

    """

    # check for required attribute before proceeding

    required_attrs = {
        CPR_GRAPH_EDGES.SC_DEGREE,
        CPR_GRAPH_EDGES.SC_CHILDREN,
        CPR_GRAPH_EDGES.SC_PARENTS,
        CPR_GRAPH_EDGES.SPECIES_TYPE,
    }

    missing_required_attrs = required_attrs.difference(set(cpr_graph.es.attributes()))
    if len(missing_required_attrs) != 0:
        raise ValueError(
            f"model is missing {len(missing_required_attrs)} required attributes: {', '.join(missing_required_attrs)}"
        )

    if base_score < 0:
        raise ValueError(f"base_score was {base_score} and must be non-negative")
    if protein_multiplier > unknown_multiplier:
        raise ValueError(
            f"protein_multiplier was {protein_multiplier} and unknown_multiplier "
            f"was {unknown_multiplier}. unknown_multiplier must be greater than "
            "protein_multiplier"
        )
    if metabolite_multiplier > unknown_multiplier:
        raise ValueError(
            f"protein_multiplier was {metabolite_multiplier} and unknown_multiplier "
            f"was {unknown_multiplier}. unknown_multiplier must be greater than "
            "protein_multiplier"
        )

    # create a new weight variable

    weight_table = pd.DataFrame(
        {
            CPR_GRAPH_EDGES.SC_DEGREE: cpr_graph.es[CPR_GRAPH_EDGES.SC_DEGREE],
            CPR_GRAPH_EDGES.SC_CHILDREN: cpr_graph.es[CPR_GRAPH_EDGES.SC_CHILDREN],
            CPR_GRAPH_EDGES.SC_PARENTS: cpr_graph.es[CPR_GRAPH_EDGES.SC_PARENTS],
            CPR_GRAPH_EDGES.SPECIES_TYPE: cpr_graph.es[CPR_GRAPH_EDGES.SPECIES_TYPE],
        }
    )

    lookup_multiplier_dict = {
        "protein": protein_multiplier,
        "metabolite": metabolite_multiplier,
        "unknown": unknown_multiplier,
    }
    weight_table["multiplier"] = weight_table["species_type"].map(
        lookup_multiplier_dict
    )

    # calculate mean degree
    # since topology weights will differ based on the structure of the network
    # and it would be nice to have a consistent notion of edge weights and path weights
    # for interpretability and filtering, we can rescale topology weights by the
    # average degree of nodes
    if scale_multiplier_by_meandegree:
        mean_degree = len(cpr_graph.es) / len(cpr_graph.vs)
        if not cpr_graph.is_directed():
            # for a directed network in- and out-degree are separately treated while
            # an undirected network's degree will be the sum of these two measures.
            mean_degree = mean_degree * 2

        weight_table["multiplier"] = weight_table["multiplier"] / mean_degree

    if cpr_graph.is_directed():
        weight_table["connection_weight"] = weight_table[CPR_GRAPH_EDGES.SC_CHILDREN]
    else:
        weight_table["connection_weight"] = weight_table[CPR_GRAPH_EDGES.SC_DEGREE]

    # weight traveling through a species based on
    # - a constant
    # - how plausibly that species type mediates a change
    # - the number of connections that the node can bridge to
    weight_table["topo_weights"] = [
        base_score + (x * y)
        for x, y in zip(weight_table["multiplier"], weight_table["connection_weight"])
    ]
    cpr_graph.es["topo_weights"] = weight_table["topo_weights"]

    # if directed and we want to use travel upstream define a corresponding weighting scheme
    if cpr_graph.is_directed():
        weight_table["upstream_topo_weights"] = [
            base_score + (x * y)
            for x, y in zip(weight_table["multiplier"], weight_table["sc_parents"])
        ]
        cpr_graph.es["upstream_topo_weights"] = weight_table["upstream_topo_weights"]

    return cpr_graph


def _validate_entity_attrs(
    entity_attrs: dict,
    validate_transformations: bool = True,
    custom_transformations: Optional[dict] = None,
) -> None:
    """Validate that graph attributes are a valid format."""

    assert isinstance(entity_attrs, dict)
    for v in entity_attrs.values():
        # check structure against pydantic config
        entity_attrs = _EntityAttrValidator(**v).model_dump()

        if validate_transformations:
            trans_name = v["trans"]
            valid_trans = set(DEFINED_WEIGHT_TRANSFORMATION.keys())
            if custom_transformations:
                valid_trans = valid_trans.union(set(custom_transformations.keys()))
            if trans_name not in valid_trans:
                raise ValueError(
                    f"transformation {trans_name} was not defined as an alias in "
                    "DEFINED_WEIGHT_TRANSFORMATION or custom_transformations. The defined transformations "
                    f"are {', '.join(valid_trans)}"
                )

    return None


class _EntityAttrValidator(BaseModel):
    table: str
    variable: str
    trans: Optional[str] = DEFAULT_WT_TRANS



"""Module to contain all constants used for representing and working with networks"""

from __future__ import annotations

from types import SimpleNamespace

from napistu.constants import SBML_DFS
from napistu.constants import SBOTERM_NAMES

CPR_GRAPH_NODES = SimpleNamespace(NAME="name")

CPR_GRAPH_EDGES = SimpleNamespace(
    DIRECTED="directed",
    FROM="from",
    R_ID=SBML_DFS.R_ID,
    R_ISREVERSIBLE=SBML_DFS.R_ISREVERSIBLE,
    SBO_TERM=SBML_DFS.SBO_TERM,
    SBO_NAME="sbo_name",
    SC_DEGREE="sc_degree",
    SC_PARENTS="sc_parents",
    SC_CHILDREN="sc_children",
    SPECIES_TYPE="species_type",
    STOICHIOMETRY=SBML_DFS.STOICHIOMETRY,
    TO="to",
    UPSTREAM_WEIGHTS="upstream_weights",
    WEIGHTS="weights",
)

# variables which should be in cpr graph's edges
CPR_GRAPH_REQUIRED_EDGE_VARS = {
    CPR_GRAPH_EDGES.FROM,
    CPR_GRAPH_EDGES.TO,
    CPR_GRAPH_EDGES.SBO_TERM,
    CPR_GRAPH_EDGES.STOICHIOMETRY,
    CPR_GRAPH_EDGES.SC_PARENTS,
    CPR_GRAPH_EDGES.SC_CHILDREN,
}

# nomenclature for individual fields

CPR_GRAPH_NODE_TYPES = SimpleNamespace(REACTION="reaction", SPECIES="species")

VALID_CPR_GRAPH_NODE_TYPES = [
    CPR_GRAPH_NODE_TYPES.REACTION,
    CPR_GRAPH_NODE_TYPES.SPECIES,
]

CPR_GRAPH_EDGE_DIRECTIONS = SimpleNamespace(
    FORWARD="forward", REVERSE="reverse", UNDIRECTED="undirected"
)

# network-level nomenclature

CPR_GRAPH_TYPES = SimpleNamespace(
    BIPARTITE="bipartite", REGULATORY="regulatory", SURROGATE="surrogate"
)

VALID_CPR_GRAPH_TYPES = [
    CPR_GRAPH_TYPES.BIPARTITE,
    CPR_GRAPH_TYPES.REGULATORY,
    CPR_GRAPH_TYPES.SURROGATE,
]

CPR_WEIGHTING_STRATEGIES = SimpleNamespace(
    CALIBRATED="calibrated", MIXED="mixed", TOPOLOGY="topology", UNWEIGHTED="unweighted"
)

VALID_WEIGHTING_STRATEGIES = [
    CPR_WEIGHTING_STRATEGIES.UNWEIGHTED,
    CPR_WEIGHTING_STRATEGIES.TOPOLOGY,
    CPR_WEIGHTING_STRATEGIES.MIXED,
    CPR_WEIGHTING_STRATEGIES.CALIBRATED,
]

# the regulatory graph defines a hierarchy of upstream and downstream
# entities in a reaction
# modifier/stimulator/inhibitor -> catalyst -> reactant -> reaction -> product

REGULATORY_GRAPH_HIERARCHY = [
    [SBOTERM_NAMES.MODIFIER, SBOTERM_NAMES.STIMULATOR, SBOTERM_NAMES.INHIBITOR],
    [SBOTERM_NAMES.CATALYST],
    [SBOTERM_NAMES.REACTANT],
    [CPR_GRAPH_NODE_TYPES.REACTION],
    # normally we don't expect interactors to be defined because they are handled by
    # net_create._format_interactors_for_regulatory_graph() but include them here
    # until Issue #102 is solved
    [SBOTERM_NAMES.INTERACTOR],
    [SBOTERM_NAMES.PRODUCT],
]

# an alternative layout to regulatory where enyzmes are downstream of substrates.
# this doesn't make much sense from a regulatory perspective because
# enzymes modify substrates not the other way around. but, its what one might
# expect if catalysts are a surrogate for reactions as is the case for metabolic
# network layouts

SURROGATE_GRAPH_HIERARCHY = [
    [SBOTERM_NAMES.MODIFIER, SBOTERM_NAMES.STIMULATOR, SBOTERM_NAMES.INHIBITOR],
    [SBOTERM_NAMES.REACTANT],
    [SBOTERM_NAMES.CATALYST],
    [CPR_GRAPH_NODE_TYPES.REACTION],
    # normally we don't expect interactors to be defined because they are handled by
    # net_create._format_interactors_for_regulatory_graph() but include them here
    # until Issue #102 is solved
    [SBOTERM_NAMES.INTERACTOR],
    [SBOTERM_NAMES.PRODUCT],
]

NEIGHBORHOOD_NETWORK_TYPES = SimpleNamespace(
    DOWNSTREAM="downstream", HOURGLASS="hourglass", UPSTREAM="upstream"
)

VALID_NEIGHBORHOOD_NETWORK_TYPES = [
    NEIGHBORHOOD_NETWORK_TYPES.DOWNSTREAM,
    NEIGHBORHOOD_NETWORK_TYPES.HOURGLASS,
    NEIGHBORHOOD_NETWORK_TYPES.UPSTREAM,
]



from __future__ import annotations

import pytest

import igraph as ig
import numpy as np
import pandas as pd
from napistu.network import net_utils
from napistu.network import net_create


def test_safe_fill():
    safe_fill_test = ["a_very_long stringggg", ""]
    assert [net_utils.safe_fill(x) for x in safe_fill_test] == [
        "a_very_long\nstringggg",
        "",
    ]


def test_cpr_graph_to_pandas_dfs():
    graph_data = [
        (0, 1),
        (0, 2),
        (2, 3),
        (3, 4),
        (4, 2),
        (2, 5),
        (5, 0),
        (6, 3),
        (5, 6),
    ]

    g = ig.Graph(graph_data, directed=True)
    vs, es = net_utils.cpr_graph_to_pandas_dfs(g)

    assert all(vs["index"] == list(range(0, 7)))
    assert (
        pd.DataFrame(graph_data)
        .rename({0: "source", 1: "target"}, axis=1)
        .sort_values(["source", "target"])
        .equals(es.sort_values(["source", "target"]))
    )


def test_validate_graph_attributes(sbml_dfs):

    cpr_graph = net_create.process_cpr_graph(
        sbml_dfs, directed=True, weighting_strategy="topology"
    )

    assert (
        net_utils._validate_edge_attributes(cpr_graph, ["weights", "upstream_weights"])
        is None
    )
    assert net_utils._validate_vertex_attributes(cpr_graph, "node_type") is None
    with pytest.raises(ValueError):
        net_utils._validate_vertex_attributes(cpr_graph, "baz")


def test_pluck_entity_data_species_identity(sbml_dfs):
    # Take first 10 species IDs
    species_ids = sbml_dfs.species.index[:10]
    # Create mock data with explicit dtype to ensure cross-platform consistency
    # Fix for issue-42: Use explicit dtypes to avoid platform-specific dtype differences
    # between Windows (int32) and macOS/Linux (int64)
    mock_df = pd.DataFrame(
        {
            "string_col": [f"str_{i}" for i in range(10)],
            "mixed_col": np.arange(-5, 5, dtype=np.int64),  # Explicitly use int64
            "ones_col": np.ones(10, dtype=np.float64),  # Explicitly use float64
            "squared_col": np.arange(10, dtype=np.int64),  # Explicitly use int64
        },
        index=species_ids,
    )
    # Assign to species_data
    sbml_dfs.species_data["mock_table"] = mock_df

    # Custom transformation: square
    def square(x):
        return x**2

    custom_transformations = {"square": square}
    # Create graph_attrs for species
    graph_attrs = {
        "species": {
            "string_col": {
                "table": "mock_table",
                "variable": "string_col",
                "trans": "identity",
            },
            "mixed_col": {
                "table": "mock_table",
                "variable": "mixed_col",
                "trans": "identity",
            },
            "ones_col": {
                "table": "mock_table",
                "variable": "ones_col",
                "trans": "identity",
            },
            "squared_col": {
                "table": "mock_table",
                "variable": "squared_col",
                "trans": "square",
            },
        }
    }
    # Call pluck_entity_data with custom transformation
    result = net_create.pluck_entity_data(
        sbml_dfs, graph_attrs, "species", custom_transformations=custom_transformations
    )
    # Check output
    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == {"string_col", "mixed_col", "ones_col", "squared_col"}
    assert list(result.index) == list(species_ids)
    # Check values
    pd.testing.assert_series_equal(result["string_col"], mock_df["string_col"])
    pd.testing.assert_series_equal(result["mixed_col"], mock_df["mixed_col"])
    pd.testing.assert_series_equal(result["ones_col"], mock_df["ones_col"])
    pd.testing.assert_series_equal(
        result["squared_col"], mock_df["squared_col"].apply(square)
    )


def test_pluck_entity_data_missing_species_key(sbml_dfs):
    # graph_attrs does not contain 'species' key
    graph_attrs = {}
    result = net_create.pluck_entity_data(sbml_dfs, graph_attrs, "species")
    assert result is None


def test_pluck_entity_data_empty_species_dict(sbml_dfs):
    # graph_attrs contains 'species' key but value is empty dict
    graph_attrs = {"species": {}}
    result = net_create.pluck_entity_data(sbml_dfs, graph_attrs, "species")
    assert result is None


################################################
# __main__
################################################

if __name__ == "__main__":
    test_safe_fill()
    test_cpr_graph_to_pandas_dfs()
    test_validate_graph_attributes()
    test_pluck_entity_data_species_identity()
    test_pluck_entity_data_missing_species_key()
    test_pluck_entity_data_empty_species_dict()



from __future__ import annotations

import logging
from typing import Optional, Union, Set, Dict, List

import igraph as ig
import numpy as np
import pandas as pd

from napistu import identifiers
from napistu import sbml_dfs_core
from napistu import utils
from napistu.constants import SBML_DFS
from napistu.constants import CPR_EDGELIST
from napistu.constants import CPR_EDGELIST_REQ_VARS
from napistu.constants import FEATURE_ID_VAR_DEFAULT
from napistu.constants import RESOLVE_MATCHES_AGGREGATORS
from napistu.constants import RESOLVE_MATCHES_TMP_WEIGHT_COL
from napistu.constants import IDENTIFIERS
from napistu.constants import IDENTIFIER_EDGELIST_REQ_VARS
from napistu.constants import ONTOLOGIES_LIST
from napistu.network.constants import CPR_GRAPH_EDGES
from napistu.network import paths

logger = logging.getLogger(__name__)


def bind_wide_results(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    results_df: pd.DataFrame,
    results_name: str,
    ontologies: Optional[Union[Set[str], Dict[str, str]]] = None,
    dogmatic: bool = False,
    species_identifiers: Optional[pd.DataFrame] = None,
    feature_id_var: str = FEATURE_ID_VAR_DEFAULT,
    numeric_agg: str = RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN,
    keep_id_col: bool = True,
    verbose: bool = False,
) -> sbml_dfs_core.SBML_dfs:
    """
    Binds wide results to a sbml_dfs object.

    Take a table with molecular species-level attributes tied to systematic identifiers and match them to an sbml_dfs_model transferring these attributes to species_data

    Parameters
    ----------
    sbml_dfs : sbml_dfs_core.SBML_dfs
        The sbml_dfs object to bind the results to.
    results_df : pd.DataFrame
        The table containing the results to bind.
    results_name : str
        The name of the results to bind.
    ontologies : Optional[Union[Set[str], Dict[str, str]]], default=None
        Either:
        - Set of columns to treat as ontologies (these should be entries in ONTOLOGIES_LIST )
        - Dict mapping wide column names to ontology names in the ONTOLOGIES_LIST controlled vocabulary
        - None to automatically detect valid ontology columns based on ONTOLOGIES_LIST
    dogmatic : bool
        Whether to respect differences between genes, transcripts, and proteins (True) or ignore them (False).
    species_identifiers : Optional[pd.DataFrame]
        Systematic identifiers for the molecular species "sbml_dfs". If None this will be generate on-the-fly.
    feature_id_var : str
        The name of the column in the results_df that contains the feature identifiers. If this does not exist it will be created.
    numeric_agg : str
        The aggregation method to use for resolving degeneracy.
    keep_id_col : bool
        Whether to keep the identifier column in the results_df.
    verbose : bool
        Whether to log cases of 1-to-many and many-to-one mapping and to indicate the behavior for resolving degeneracy

    Returns
    -------
    sbml_dfs : sbml_dfs_core.SBML_dfs
        The sbml_dfs object with the results bound.
    """

    species_identifiers = identifiers._prepare_species_identifiers(
        sbml_dfs, dogmatic=dogmatic, species_identifiers=species_identifiers
    )

    # match
    matched_s_ids_from_wide = match_features_to_wide_pathway_species(
        results_df,
        species_identifiers,
        ontologies=ontologies,
        feature_id_var=feature_id_var,
        verbose=verbose,
    )

    disambiguated_matches = resolve_matches(
        matched_data=matched_s_ids_from_wide,
        feature_id_var=feature_id_var,
        numeric_agg=numeric_agg,
        keep_id_col=keep_id_col,
    )

    clean_species_data = utils.drop_extra_cols(
        results_df, disambiguated_matches, always_include=[feature_id_var]
    )

    sbml_dfs.add_species_data(results_name, clean_species_data)

    return sbml_dfs


def features_to_pathway_species(
    feature_identifiers: pd.DataFrame,
    species_identifiers: pd.DataFrame,
    ontologies: set,
    feature_identifiers_var: str = IDENTIFIERS.IDENTIFIER,
    feature_id_var: str = FEATURE_ID_VAR_DEFAULT,
    expand_identifiers: bool = False,
    identifier_delimiter: str = "/",
    verbose: bool = False,
) -> pd.DataFrame:
    """
    Features to Pathway Species

    Match a table of molecular species to their corresponding species in a pathway representation.

    Parameters:
    feature_identifiers: pd.DataFrame
        pd.Dataframe containing a "feature_identifiers_var" variable used to match entries
    species_identifiers: pd.DataFrame
        A table of molecular species identifiers produced from sbml_dfs.get_identifiers("species")
        generally using sbml_dfs_core.export_sbml_dfs()
    ontologies: set
        A set of ontologies used to match features to pathway species
    feature_identifiers_var: str
        Variable in "feature_identifiers" containing identifiers
    expand_identifiers: bool, default=False
        If True, split identifiers in feature_identifiers_var by identifier_delimiter and explode into multiple rows
    identifier_delimiter: str, default="/"
        Delimiter to use for splitting identifiers if expand_identifiers is True
    verbose: bool, default=False
        If True, log mapping statistics at the end of the function

    Returns:
    pathway_species: pd.DataFrame
        species_identifiers joined to feature_identifiers based on shared identifiers
    """

    # Check for identifier column
    if feature_identifiers_var not in feature_identifiers.columns.to_list():
        raise ValueError(
            f"{feature_identifiers_var} must be a variable in 'feature_identifiers', "
            f"possible variables are {', '.join(feature_identifiers.columns.tolist())}"
        )

    # Respect or create feature_id column
    feature_identifiers = _ensure_feature_id_var(feature_identifiers, feature_id_var)

    # Optionally expand identifiers into multiple rows
    if expand_identifiers:
        # Count the number of expansions by counting delimiters
        n_expansions = (
            feature_identifiers[feature_identifiers_var]
            .astype(str)
            .str.count(identifier_delimiter)
            .sum()
        )
        if n_expansions > 0:
            logger.info(
                f"Expanding identifiers: {n_expansions} delimiters found in '{feature_identifiers_var}', will expand to more rows."
            )

        # Split, strip whitespace, and explode
        feature_identifiers = feature_identifiers.copy()
        feature_identifiers[feature_identifiers_var] = (
            feature_identifiers[feature_identifiers_var]
            .astype(str)
            .str.split(identifier_delimiter)
            .apply(lambda lst: [x.strip() for x in lst])
        )
        feature_identifiers = feature_identifiers.explode(
            feature_identifiers_var, ignore_index=True
        )

    # check identifiers table
    identifiers._check_species_identifiers_table(species_identifiers)

    available_ontologies = set(species_identifiers[IDENTIFIERS.ONTOLOGY].tolist())
    unavailable_ontologies = ontologies.difference(available_ontologies)

    # no ontologies present
    if len(unavailable_ontologies) == len(ontologies):
        raise ValueError(
            f"None of the requested ontologies ({', '.join(ontologies)}) "
            "were used to annotate pathway species. Available ontologies are: "
            f"{', '.join(available_ontologies)}"
        )

    # 1+ desired ontologies are not present
    if len(unavailable_ontologies) > 0:
        raise ValueError(
            f"Some of the requested ontologies ({', '.join(unavailable_ontologies)}) "
            "were NOT used to annotate pathway species. Available ontologies are: "
            f"{', '.join(available_ontologies)}"
        )

    relevant_identifiers = species_identifiers[
        species_identifiers[IDENTIFIERS.ONTOLOGY].isin(ontologies)
    ]

    # map features to pathway species
    pathway_species = feature_identifiers.merge(
        relevant_identifiers,
        left_on=feature_identifiers_var,
        right_on=IDENTIFIERS.IDENTIFIER,
    )

    if pathway_species.shape[0] == 0:
        logger.warning(
            "None of the provided species identifiers matched entries of the pathway; returning None"
        )
        None

    # report the fraction of unmapped species
    if verbose:
        _log_feature_species_mapping_stats(pathway_species, feature_id_var)

    return pathway_species


def edgelist_to_pathway_species(
    formatted_edgelist: pd.DataFrame,
    species_identifiers: pd.DataFrame,
    ontologies: set,
    feature_id_var: str = FEATURE_ID_VAR_DEFAULT,
    verbose: bool = False,
) -> pd.DataFrame:
    """
    Edgelist to Pathway Species

    Match an edgelist of molecular species pairs to their corresponding species in a pathway representation.

    Parameters:
    formatted_edgelist: pd.DataFrame
        pd.Dataframe containing a "identifier_upstream" and "identifier_downstream" variables used to to match entries
    species_identifiers: pd.DataFrame
        A table of molecular species identifiers produced from sbml_dfs.get_identifiers("species") generally using
        sbml_dfs_core.export_sbml_dfs()
    ontologies: set
        A set of ontologies used to match features to pathway species
    feature_id_var: str, default=FEATURE_ID_VAR_DEFAULT
        Variable in "formatted_edgelist" containing feature ids
    verbose: bool, default=False
        Whether to print verbose output

    Returns:
    edges_on_pathway: pd.DataFrame
        formatted_edgelist with upstream features mapped
        to "s_id_upstream" and downstream species mapped
        to "s_id_downstream"
    """

    required_vars_distinct_features = {
        CPR_EDGELIST.IDENTIFIER_UPSTREAM,
        CPR_EDGELIST.IDENTIFIER_DOWNSTREAM,
    }
    missing_required_vars_distinct_features = (
        required_vars_distinct_features.difference(
            set(formatted_edgelist.columns.tolist())
        )
    )

    if len(missing_required_vars_distinct_features) > 0:
        raise ValueError(
            f"{len(missing_required_vars_distinct_features)} required variables were "
            "missing from 'formatted_edgelist': "
            f"{', '.join(missing_required_vars_distinct_features)}"
        )

    # define all distinct identifiers in edgelist
    distinct_identifiers = (
        pd.concat(
            [
                formatted_edgelist[CPR_EDGELIST.IDENTIFIER_UPSTREAM],
                formatted_edgelist[CPR_EDGELIST.IDENTIFIER_DOWNSTREAM],
            ]
        )
        .drop_duplicates()
        .reset_index(drop=True)
        .to_frame()
        .rename({0: feature_id_var}, axis=1)
    )

    # merge edgelist identifiers with pathway identifiers to map s_ids to identifiers
    features_on_pathway = features_to_pathway_species(
        feature_identifiers=distinct_identifiers,
        species_identifiers=species_identifiers,
        ontologies=ontologies,
        feature_identifiers_var=feature_id_var,
        verbose=verbose,
    )

    # add s_ids of both upstream and downstream edges to pathway
    edges_on_pathway = formatted_edgelist.merge(
        features_on_pathway[[SBML_DFS.S_ID, IDENTIFIERS.IDENTIFIER]].rename(
            {
                SBML_DFS.S_ID: CPR_EDGELIST.S_ID_UPSTREAM,
                IDENTIFIERS.IDENTIFIER: CPR_EDGELIST.IDENTIFIER_UPSTREAM,
            },
            axis=1,
        )
    ).merge(
        features_on_pathway[[SBML_DFS.S_ID, IDENTIFIERS.IDENTIFIER]].rename(
            {
                SBML_DFS.S_ID: CPR_EDGELIST.S_ID_DOWNSTREAM,
                IDENTIFIERS.IDENTIFIER: CPR_EDGELIST.IDENTIFIER_DOWNSTREAM,
            },
            axis=1,
        )
    )

    return edges_on_pathway


def match_features_to_wide_pathway_species(
    wide_df: pd.DataFrame,
    species_identifiers: pd.DataFrame,
    ontologies: Optional[Union[Set[str], Dict[str, str]]] = None,
    feature_identifiers_var: str = IDENTIFIERS.IDENTIFIER,
    feature_id_var: str = FEATURE_ID_VAR_DEFAULT,
    verbose: bool = False,
) -> pd.DataFrame:
    """
    Convert a wide-format DataFrame with multiple ontology columns to long format,
    and match features to pathway species by ontology and identifier.

    Parameters
    ----------
    wide_df : pd.DataFrame
        DataFrame with ontology identifier columns and any number of results columns.
        All non-ontology columns are treated as results.
    species_identifiers : pd.DataFrame
        DataFrame as required by features_to_pathway_species
    ontologies : Optional[Union[Set[str], Dict[str, str]]], default=None
        Either:
        - Set of columns to treat as ontologies (these should be entries in ONTOLOGIES_LIST )
        - Dict mapping wide column names to ontology names in the ONTOLOGIES_LIST controlled vocabulary
        - None to automatically detect valid ontology columns based on ONTOLOGIES_LIST
    feature_identifiers_var : str, default="identifier"
        Name for the identifier column in the long format
    feature_id_var: str, default=FEATURE_ID_VAR_DEFAULT
        Name for the feature id column in the long format
    verbose : bool, default=False
        Whether to print verbose output

    Returns
    -------
    pd.DataFrame
        Output of match_by_ontology_and_identifier

    Examples
    --------
    >>> # Example with auto-detected ontology columns and multiple results
    >>> wide_df = pd.DataFrame({
    ...     'uniprot': ['P12345', 'Q67890'],
    ...     'chebi': ['15377', '16810'],
    ...     'log2fc': [1.0, 2.0],
    ...     'pvalue': [0.01, 0.05]
    ... })
    >>> result = match_features_to_wide_pathway_species(
    ...     wide_df=wide_df,
    ...     species_identifiers=species_identifiers
    ... )

    >>> # Example with custom ontology mapping
    >>> wide_df = pd.DataFrame({
    ...     'protein_id': ['P12345', 'Q67890'],
    ...     'compound_id': ['15377', '16810'],
    ...     'expression': [1.0, 2.0],
    ...     'confidence': [0.8, 0.9]
    ... })
    >>> result = match_features_to_wide_pathway_species(
    ...     wide_df=wide_df,
    ...     species_identifiers=species_identifiers,
    ...     ontologies={'protein_id': 'uniprot', 'compound_id': 'chebi'}
    ... )
    """
    # Make a copy to avoid modifying the input
    wide_df = wide_df.copy()

    # Validate ontologies and get the set of ontology columns
    ontology_cols = _validate_wide_ontologies(wide_df, ontologies)
    melt_cols = list(ontology_cols)

    # Apply renaming if a mapping is provided
    if isinstance(ontologies, dict):
        wide_df = wide_df.rename(columns=ontologies)

    # Ensure feature_id column exists
    wide_df = _ensure_feature_id_var(wide_df, feature_id_var)

    # All non-ontology columns are treated as results
    results_cols = list(set(wide_df.columns) - set(melt_cols))
    if not results_cols:
        raise ValueError("No results columns found in DataFrame")

    logger.info(f"Using columns as results: {results_cols}")

    # Melt ontology columns to long format, keeping all results columns
    long_df = wide_df.melt(
        id_vars=results_cols,
        value_vars=melt_cols,
        var_name=IDENTIFIERS.ONTOLOGY,
        value_name=feature_identifiers_var,
    ).dropna(subset=[feature_identifiers_var])

    logger.debug(f"Final long format shape: {long_df.shape}")

    # Call the matching function with the validated ontologies
    out = match_by_ontology_and_identifier(
        feature_identifiers=long_df,
        species_identifiers=species_identifiers,
        ontologies=ontology_cols,
        feature_identifiers_var=feature_identifiers_var,
    )

    if verbose:
        _log_feature_species_mapping_stats(out, feature_id_var)

    return out


def match_by_ontology_and_identifier(
    feature_identifiers: pd.DataFrame,
    species_identifiers: pd.DataFrame,
    ontologies: Union[str, Set[str], List[str]],
    feature_identifiers_var: str = IDENTIFIERS.IDENTIFIER,
    verbose: bool = False,
) -> pd.DataFrame:
    """
    Match features to pathway species based on both ontology and identifier matches.
    Performs separate matching for each ontology and concatenates the results.

    Parameters
    ----------
    feature_identifiers : pd.DataFrame
        DataFrame containing feature identifiers and results.
        Must have columns [ontology, feature_identifiers_var, results]
    species_identifiers : pd.DataFrame
        DataFrame containing species identifiers from pathway.
        Must have columns [ontology, identifier]
    ontologies : Union[str, Set[str], List[str]]
        Ontologies to match on. Can be:
        - A single ontology string
        - A set of ontology strings
        - A list of ontology strings
    feature_identifiers_var : str, default="identifier"
        Name of the identifier column in feature_identifiers
    verbose : bool, default=False
        Whether to print verbose output

    Returns
    -------
    pd.DataFrame
        Concatenated results of matching for each ontology.
        Contains all columns from features_to_pathway_species()

    Examples
    --------
    >>> # Match using a single ontology
    >>> result = match_by_ontology_and_identifier(
    ...     feature_identifiers=features_df,
    ...     species_identifiers=species_df,
    ...     ontologies="uniprot"
    ... )

    >>> # Match using multiple ontologies
    >>> result = match_by_ontology_and_identifier(
    ...     feature_identifiers=features_df,
    ...     species_identifiers=species_df,
    ...     ontologies={"uniprot", "chebi"}
    ... )
    """
    # Convert string to set for consistent handling
    if isinstance(ontologies, str):
        ontologies = {ontologies}
    elif isinstance(ontologies, list):
        ontologies = set(ontologies)

    # Validate ontologies
    invalid_onts = ontologies - set(ONTOLOGIES_LIST)
    if invalid_onts:
        raise ValueError(
            f"Invalid ontologies specified: {invalid_onts}. Must be one of: {ONTOLOGIES_LIST}"
        )

    # Initialize list to store results
    matched_dfs = []

    # Process each ontology separately
    for ont in ontologies:
        # Filter feature identifiers to current ontology and drop ontology column
        ont_features = (
            feature_identifiers[feature_identifiers[IDENTIFIERS.ONTOLOGY] == ont]
            .drop(columns=[IDENTIFIERS.ONTOLOGY])
            .copy()
        )

        if ont_features.empty:
            logger.warning(f"No features found for ontology: {ont}")
            continue

        # Filter species identifiers to current ontology
        ont_species = species_identifiers[
            species_identifiers[IDENTIFIERS.ONTOLOGY] == ont
        ].copy()

        if ont_species.empty:
            logger.warning(f"No species found for ontology: {ont}")
            continue

        logger.debug(
            f"Matching {len(ont_features)} features to {len(ont_species)} species for ontology {ont}"
        )

        # Match features to species for this ontology
        matched = features_to_pathway_species(
            feature_identifiers=ont_features,
            species_identifiers=ont_species,
            ontologies={ont},
            feature_identifiers_var=feature_identifiers_var,
            verbose=verbose,
        )

        if matched.empty:
            logger.warning(f"No matches found for ontology: {ont}")
            continue

        matched_dfs.append(matched)

    if not matched_dfs:
        logger.warning("No matches found for any ontology")
        return pd.DataFrame()  # Return empty DataFrame with correct columns

    # Combine results from all ontologies
    result = pd.concat(matched_dfs, axis=0, ignore_index=True)

    logger.info(
        f"Found {len(result)} total matches across {len(matched_dfs)} ontologies"
    )

    return result


def resolve_matches(
    matched_data: pd.DataFrame,
    feature_id_var: str = FEATURE_ID_VAR_DEFAULT,
    index_col: str = SBML_DFS.S_ID,
    numeric_agg: str = RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN,
    keep_id_col: bool = True,
) -> pd.DataFrame:
    """
    Resolve many-to-1 and 1-to-many matches in matched data.

    Parameters
    ----------
    matched_data : pd.DataFrame
        DataFrame containing matched data with columns:
        - feature_id_var: identifier column (e.g. feature_id)
        - index_col: index column (e.g. s_id)
        - other columns: data columns to be aggregated
    feature_id_var : str, default="feature_id"
        Name of the identifier column
    index_col : str, default="s_id"
        Name of the column to use as index
    numeric_agg : str, default="weighted_mean"
        Method to aggregate numeric columns:
        - "weighted_mean": weighted by inverse of feature_id frequency (default)
        - "mean": simple arithmetic mean
        - "first": first value after sorting by feature_id_var (requires feature_id_var)
        - "max": maximum value
    keep_id_col : bool, default=True
        Whether to keep and rollup the feature_id_var in the output.
        If False, feature_id_var will be dropped from the output.

    Returns
    -------
    pd.DataFrame
        DataFrame with resolved matches:
        - Many-to-1: numeric columns are aggregated using specified method
        - 1-to-many: adds a count column showing number of matches
        - Index is set to index_col and named accordingly

    Raises
    ------
    KeyError
        If feature_id_var is not present in the DataFrame
    TypeError
        If DataFrame contains unsupported data types (boolean or datetime)
    """
    # Make a copy to avoid modifying input
    df = matched_data.copy()

    # Check for unsupported data types
    unsupported_dtypes = df.select_dtypes(include=["bool", "datetime64"]).columns
    if not unsupported_dtypes.empty:
        raise TypeError(
            f"Unsupported data types found in columns: {list(unsupported_dtypes)}. "
            "Boolean and datetime columns are not supported."
        )

    # Always require feature_id_var
    if feature_id_var not in df.columns:
        raise KeyError(feature_id_var)

    # Deduplicate by feature_id within each s_id using groupby and first BEFORE any further processing
    df = df.groupby([index_col, feature_id_var], sort=False).first().reset_index()

    # Use a unique temporary column name for weights
    if RESOLVE_MATCHES_TMP_WEIGHT_COL in df.columns:
        raise ValueError(
            f"Temporary weight column name '{RESOLVE_MATCHES_TMP_WEIGHT_COL}' already exists in the input data. Please rename or remove this column and try again."
        )

    # Calculate weights if needed (after deduplication!)
    if numeric_agg == RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN:
        feature_counts = df[feature_id_var].value_counts()
        df[RESOLVE_MATCHES_TMP_WEIGHT_COL] = (
            1 / feature_counts[df[feature_id_var]].values
        )

    # Set index for grouping
    df = df.set_index(index_col)

    # Use utility to split columns
    always_non_numeric = [feature_id_var] if keep_id_col else []
    numeric_cols, non_numeric_cols = _split_numeric_non_numeric_columns(
        df, always_non_numeric=always_non_numeric
    )

    # Get aggregator function
    numeric_aggregator = _get_numeric_aggregator(
        method=numeric_agg, feature_id_var=feature_id_var
    )
    resolved = _aggregate_grouped_columns(
        df,
        numeric_cols,
        non_numeric_cols,
        numeric_aggregator,
        feature_id_var=feature_id_var,
        numeric_agg=numeric_agg,
    )
    # Add count of matches per feature_id
    match_counts = matched_data.groupby(index_col)[feature_id_var].nunique()
    resolved[f"{feature_id_var}_match_count"] = match_counts

    # Drop feature_id_var if not keeping it
    if not keep_id_col and feature_id_var in resolved.columns:
        resolved = resolved.drop(columns=[feature_id_var])

    # Ensure index is named consistently
    resolved.index.name = index_col

    return resolved


def edgelist_to_scids(
    formatted_edgelist: pd.DataFrame,
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    species_identifiers: pd.DataFrame,
    ontologies: set,
):
    """

    Edgelist to Compartmentalized Species IDds

    Map an edgelist of possible mechanistic interactions onto a
    pathadex pathway

    Parameters:
    formatted_edgelist: pd.DataFrame
        pd.Dataframe containing a "identifier_upstream" and
        "identifier_downstream" variables used to to match entries
    sbml_dfs: sbml_dfs_core.SBML_dfs
        A mechanistic model
    species_identifiers: pd.DataFrame
        A table of molecular species identifiers produced from
        sbml_dfs.get_identifiers("species") generally using sbml_dfs_core.export_sbml_dfs()
    ontologies: set
        A set of ontologies used to match features to pathway species

    Returns:
    edgelist_w_scids: pd.DataFrame
        formatted_edgelist with upstream features mapped to "sc_id_upstream" and
        downstream species mapped to "sc_id_downstream"
    """

    identifiers._check_species_identifiers_table(species_identifiers)

    # map edges onto pathway entities based on shared identifiers
    edges_on_pathway = edgelist_to_pathway_species(
        formatted_edgelist=formatted_edgelist,
        species_identifiers=species_identifiers,
        ontologies=ontologies,
    )

    # expand from s_ids to sc_ids
    s_id_pairs = edges_on_pathway[
        [CPR_EDGELIST.S_ID_UPSTREAM, CPR_EDGELIST.S_ID_DOWNSTREAM]
    ].drop_duplicates()
    sc_id_pairs = s_id_pairs.merge(
        sbml_dfs.compartmentalized_species[[SBML_DFS.S_ID]]
        .reset_index()
        .rename(
            {
                SBML_DFS.S_ID: CPR_EDGELIST.S_ID_UPSTREAM,
                SBML_DFS.SC_ID: CPR_EDGELIST.SC_ID_UPSTREAM,
            },
            axis=1,
        )
    ).merge(
        sbml_dfs.compartmentalized_species[[SBML_DFS.S_ID]]
        .reset_index()
        .rename(
            {
                SBML_DFS.S_ID: CPR_EDGELIST.S_ID_DOWNSTREAM,
                SBML_DFS.SC_ID: CPR_EDGELIST.SC_ID_DOWNSTREAM,
            },
            axis=1,
        )
    )

    # map sc_ids back to edges_on_pathway
    # join lookup table of s_id_upstream, s_id_downstream -> sc_ids
    edgelist_w_scids = edges_on_pathway.merge(sc_id_pairs)

    logger_msg = (
        f"{edgelist_w_scids.shape[0]} interactions mapped "
        "onto pairs of compartmentalized species in the mechanistic model"
    )
    if edgelist_w_scids.shape[0] == 0:
        logger.warning(logger_msg)
    else:
        logger.info(logger_msg)

    return edgelist_w_scids


def filter_to_direct_mechanistic_interactions(
    formatted_edgelist: pd.DataFrame,
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    species_identifiers: pd.DataFrame,
    ontologies: set,
) -> pd.DataFrame:
    """
    Filter to Direct Mechanistic Interactions

    Filter an edgelist to direct mechanistic interactions

    Parameters:
    formatted_edgelist: pd.DataFrame
        pd.Dataframe containing a "identifier_upstream" and "identifier_downstream" variables used to to match entries
    sbml_dfs: sbml_dfs_core.SBML_dfs
        A mechanistic model
    species_identifiers: pd.DataFrame
        A table of molecular species identifiers
        produced from sbml_dfs.get_identifiers("species") generally
        using sbml_dfs_core.export_sbml_dfs()
    ontologies: set
        A set of ontologies used to match features to pathway species

    Returns:
    edgelist_w_direct_mechanistic_interactions: pd.DataFrame
        formatted_edgelist filtered to mechanistic reactions present in the pathway representation
    """

    edgelist_w_scids = _edgelist_to_scids_if_needed(
        formatted_edgelist, sbml_dfs, species_identifiers, ontologies
    )

    # reduce to distinct sc_id pairs
    sc_id_pairs = edgelist_w_scids[list(CPR_EDGELIST_REQ_VARS)].drop_duplicates()

    # define all existing direct regulatory interactions
    pathway_interactions = pd.concat(
        [
            # pair 0 -> <0 # modifiers affect substrates
            sbml_dfs.reaction_species[
                sbml_dfs.reaction_species[SBML_DFS.STOICHIOMETRY] == 0
            ][[SBML_DFS.R_ID, SBML_DFS.SC_ID]]
            .rename({SBML_DFS.SC_ID: CPR_EDGELIST.SC_ID_UPSTREAM}, axis=1)
            .merge(
                sbml_dfs.reaction_species[
                    sbml_dfs.reaction_species[SBML_DFS.STOICHIOMETRY] < 0
                ][[SBML_DFS.R_ID, SBML_DFS.SC_ID]].rename(
                    {SBML_DFS.SC_ID: CPR_EDGELIST.SC_ID_DOWNSTREAM}, axis=1
                )
            ),
            # pair <0 -> >0 # substrates affect products
            sbml_dfs.reaction_species[
                sbml_dfs.reaction_species[SBML_DFS.STOICHIOMETRY] < 0
            ][[SBML_DFS.R_ID, SBML_DFS.SC_ID]]
            .rename({SBML_DFS.SC_ID: CPR_EDGELIST.SC_ID_UPSTREAM}, axis=1)
            .merge(
                sbml_dfs.reaction_species[
                    sbml_dfs.reaction_species[SBML_DFS.STOICHIOMETRY] > 0
                ][[SBML_DFS.R_ID, SBML_DFS.SC_ID]].rename(
                    {SBML_DFS.SC_ID: CPR_EDGELIST.SC_ID_DOWNSTREAM}, axis=1
                )
            ),
            # pair 0 -> >0 # modifiers affect products
            sbml_dfs.reaction_species[
                sbml_dfs.reaction_species[SBML_DFS.STOICHIOMETRY] == 0
            ][[SBML_DFS.R_ID, SBML_DFS.SC_ID]]
            .rename({SBML_DFS.SC_ID: CPR_EDGELIST.SC_ID_UPSTREAM}, axis=1)
            .merge(
                sbml_dfs.reaction_species[
                    sbml_dfs.reaction_species[SBML_DFS.STOICHIOMETRY] > 0
                ][[SBML_DFS.R_ID, SBML_DFS.SC_ID]].rename(
                    {SBML_DFS.SC_ID: CPR_EDGELIST.SC_ID_DOWNSTREAM}, axis=1
                )
            ),
        ]
    ).reset_index(drop=True)

    # filter pathway interactions based on matches to sc_id_pairs
    direct_edge_interactions = (
        sc_id_pairs.merge(pathway_interactions)
        .merge(
            sbml_dfs.species[SBML_DFS.S_NAME]
            .to_frame()
            .rename({SBML_DFS.S_NAME: CPR_EDGELIST.S_NAME_UPSTREAM}, axis=1),
            left_on=CPR_EDGELIST.S_ID_UPSTREAM,
            right_index=True,
            # add species metadata for matches
        )
        .merge(
            sbml_dfs.species[SBML_DFS.S_NAME]
            .to_frame()
            .rename({SBML_DFS.S_NAME: CPR_EDGELIST.S_NAME_DOWNSTREAM}, axis=1),
            left_on=CPR_EDGELIST.S_ID_DOWNSTREAM,
            right_index=True,
            # add metadata for reactions where interaction occurs
        )
        .merge(
            sbml_dfs.reactions[SBML_DFS.R_NAME].to_frame(),
            left_on=SBML_DFS.R_ID,
            right_index=True,
        )
    )

    edgelist_w_direct_mechanistic_interactions = edgelist_w_scids.merge(
        direct_edge_interactions[
            [
                CPR_EDGELIST.SC_ID_UPSTREAM,
                CPR_EDGELIST.SC_ID_DOWNSTREAM,
                SBML_DFS.R_ID,
                CPR_EDGELIST.S_NAME_UPSTREAM,
                CPR_EDGELIST.S_NAME_DOWNSTREAM,
                SBML_DFS.R_NAME,
            ]
        ]
    )

    return edgelist_w_direct_mechanistic_interactions


def filter_to_indirect_mechanistic_interactions(
    formatted_edgelist: pd.DataFrame,
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    species_identifiers: pd.DataFrame,
    cpr_graph: ig.Graph,
    ontologies: set,
    precomputed_distances=None,
    max_path_length=10,
):
    """
    Filter to Indirect Mechanistic Interactions

    Filter an edgelist to indirect mechanistic interactions.
    Indirect relationships are identified by searching a
    network for paths from an upstream species to a downstream species

    Parameters:
    formatted_edgelist: pd.DataFrame
        pd.Dataframe containing a "identifier_upstream" and
        "identifier_downstream" variables used to to match entries
    sbml_dfs: sbml_dfs_core.SBML_dfs
        A mechanistic model
    species_identifiers: pandas.DataFrame
        A table of molecular species identifiers produced from
        sbml_dfs.get_identifiers("species") generally using sbml_dfs_core.export_sbml_dfs()
    cpr_graph: igraph.Graph
        A network representation of the sbml_dfs model
    ontologies: set
        A set of ontologies used to match features to pathway species
    precomputed_distances: None or a pd.DataFrame containing path lengths and weights
        between pairs of cspecies.
    max_path_length: int
        Maximum number of steps to consider.

    Returns:
    edgelist_w_indirect_mechanistic_interactions: pd.DataFrame
        formatted_edgelist filtered to mechanistic reactions which can be described
        by an indirect mechanism. The mechanism is described by a path weight, length,
        and a vpath and epath list of vertices and edges which were traversed to create the path.
    """

    edgelist_w_scids = _edgelist_to_scids_if_needed(
        formatted_edgelist, sbml_dfs, species_identifiers, ontologies
    )

    if precomputed_distances is not None:
        # rename to match conventions in precomputed_distances
        # filter by these precomputed distances and then restore naming
        edgelist_w_scids = paths._filter_paths_by_precomputed_distances(
            edgelist_w_scids.rename(
                {
                    CPR_EDGELIST.SC_ID_UPSTREAM: CPR_EDGELIST.SC_ID_ORIGIN,
                    CPR_EDGELIST.SC_ID_DOWNSTREAM: CPR_EDGELIST.SC_ID_DEST,
                },
                axis=1,
            ),
            precomputed_distances,
        ).rename(
            {
                CPR_EDGELIST.SC_ID_ORIGIN: CPR_EDGELIST.SC_ID_UPSTREAM,
                CPR_EDGELIST.SC_ID_DEST: CPR_EDGELIST.SC_ID_DOWNSTREAM,
            },
            axis=1,
        )

    # find paths from 1 upstream to all desired downstream sc_ids
    # (this is the convention with igraph)
    indexed_origin_vertices = edgelist_w_scids.set_index(CPR_EDGELIST.SC_ID_UPSTREAM)

    # loop through upstream cspecies and find paths to all downstream species
    global_dict = dict()
    for an_origin_index in indexed_origin_vertices.index.unique():  # type: ignore
        origin_targets = indexed_origin_vertices.loc[
            an_origin_index
        ]  # type: pd.DataFrame

        # if indexing only a single entry pd.DataFrame becomes a pd.Series
        # convert back to DataFrame for consistency
        origin_targets = utils.ensure_pd_df(origin_targets)

        # log entry for debugging
        logger.debug(
            f"finding paths from {an_origin_index} to "
            f"{origin_targets.shape[0]} target vertices"
        )

        # find all paths from indexed_origin to desired destination
        shortest_paths = paths.find_shortest_reaction_paths(
            cpr_graph,
            sbml_dfs,
            origin=an_origin_index,
            # find all unique destinations (as a list for compatibility with igraph dest)
            dest=origin_targets[CPR_EDGELIST.SC_ID_DOWNSTREAM].unique().tolist(),
            weight_var=CPR_GRAPH_EDGES.WEIGHTS,
        )

        if shortest_paths is None:
            continue

        vertices, edges = shortest_paths
        indexed_edges = edges.set_index("path")
        indexed_vertices = vertices.set_index("path")

        paths_list = list()
        for ind in indexed_edges.index.unique():
            one_path = indexed_edges.loc[ind]

            # make sure that we are working with a DF
            if type(one_path) is pd.Series:
                one_path = one_path.to_frame().T

            if one_path.shape[0] > max_path_length:
                continue

            # find the destination node
            # this is annoying because if the graph is undirected
            # its not clear if the from or to edge is the actual destination
            # when taking advantage of the fact that igraph lets you
            # look up multiple destinations at once this information is lost
            ancestor_species = {an_origin_index}
            if one_path.shape[0] > 1:
                penultimate_edge = one_path.iloc[one_path.shape[0] - 2]
                ancestor_species = ancestor_species.union(
                    {
                        penultimate_edge[CPR_GRAPH_EDGES.FROM],
                        penultimate_edge[CPR_GRAPH_EDGES.TO],
                    }
                )

            terminal_edge = one_path.iloc[one_path.shape[0] - 1]
            ending_cspecies = {terminal_edge[CPR_GRAPH_EDGES.FROM], terminal_edge[CPR_GRAPH_EDGES.TO]}.difference(ancestor_species)  # type: ignore

            if len(ending_cspecies) != 1:
                raise ValueError(
                    "The terminal edge could not be determined when summarizing paths"
                )
            ending_cspecies = ending_cspecies.pop()

            path_series = pd.Series(
                {
                    CPR_GRAPH_EDGES.FROM: an_origin_index,
                    CPR_GRAPH_EDGES.TO: ending_cspecies,
                    "weight": sum(one_path[CPR_GRAPH_EDGES.WEIGHTS]),
                    "path_length": one_path.shape[0],
                    "vpath": indexed_vertices.loc[ind],
                    "epath": one_path,
                }  # type: ignore
            )  # type: pd.Series

            paths_list.append(path_series)

        if len(paths_list) > 0:
            origin_paths = pd.DataFrame(paths_list)
            global_dict[an_origin_index] = origin_paths

    if len(global_dict.keys()) == 0:
        logger.warning(
            "None of the provide molecular pairs could be mechanistically linked with a network path"
        )
        return None

    all_shortest_paths = pd.concat(global_dict.values())

    indirect_shortest_paths = edgelist_w_scids.merge(
        all_shortest_paths,
        left_on=[CPR_EDGELIST.SC_ID_UPSTREAM, CPR_EDGELIST.SC_ID_DOWNSTREAM],
        right_on=[CPR_GRAPH_EDGES.FROM, CPR_GRAPH_EDGES.TO],
    )

    return indirect_shortest_paths


def _edgelist_to_scids_if_needed(
    edgelist: pd.DataFrame,
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    species_identifiers: pd.DataFrame,
    ontologies: set,
) -> pd.DataFrame:
    """Map a set of edgelist species to cspecies or skip if cspecies were provided."""

    if utils.match_pd_vars(edgelist, CPR_EDGELIST_REQ_VARS).are_present:
        logger.info(
            f"An edgelist with {', '.join(CPR_EDGELIST_REQ_VARS)} was provided; identifier matching will be skipped"
        )
        return edgelist
    else:
        utils.match_pd_vars(edgelist, IDENTIFIER_EDGELIST_REQ_VARS).assert_present()

        identifiers._check_species_identifiers_table(species_identifiers)

        edgelist_w_scids = edgelist_to_scids(
            edgelist,
            sbml_dfs=sbml_dfs,
            species_identifiers=species_identifiers,
            ontologies=ontologies,
        )

        return edgelist_w_scids


def _validate_wide_ontologies(
    wide_df: pd.DataFrame,
    ontologies: Optional[Union[str, Set[str], Dict[str, str]]] = None,
) -> Set[str]:
    """
    Validate ontology specifications against the wide DataFrame and ONTOLOGIES_LIST.

    Parameters
    ----------
    wide_df : pd.DataFrame
        DataFrame with one column per ontology and a results column
    ontologies : Optional[Union[str, Set[str], Dict[str, str]]]
        Either:
        - String specifying a single ontology column
        - Set of columns to treat as ontologies
        - Dict mapping wide column names to ontology names
        - None to automatically detect ontology columns based on ONTOLOGIES_LIST

    Returns
    -------
    Set[str]
        Set of validated ontology names. For dictionary mappings, returns the target ontology names.

    Raises
    ------
    ValueError
        If validation fails for any ontology specification or no valid ontologies are found
    """
    # Convert string input to set
    if isinstance(ontologies, str):
        ontologies = {ontologies}

    # Get the set of ontology columns
    if isinstance(ontologies, dict):
        # Check source columns exist in DataFrame
        missing_cols = set(ontologies.keys()) - set(wide_df.columns)
        if missing_cols:
            raise ValueError(f"Source columns not found in DataFrame: {missing_cols}")
        # Validate target ontologies against ONTOLOGIES_LIST
        invalid_onts = set(ontologies.values()) - set(ONTOLOGIES_LIST)
        if invalid_onts:
            raise ValueError(
                f"Invalid ontologies in mapping: {invalid_onts}. Must be one of: {ONTOLOGIES_LIST}"
            )
        # Return target ontology names instead of source column names
        ontology_cols = set(ontologies.values())

    elif isinstance(ontologies, set):
        # Check specified columns exist in DataFrame
        missing_cols = ontologies - set(wide_df.columns)
        if missing_cols:
            raise ValueError(
                f"Specified ontology columns not found in DataFrame: {missing_cols}"
            )
        # Validate specified ontologies against ONTOLOGIES_LIST
        invalid_onts = ontologies - set(ONTOLOGIES_LIST)
        if invalid_onts:
            raise ValueError(
                f"Invalid ontologies in set: {invalid_onts}. Must be one of: {ONTOLOGIES_LIST}"
            )
        ontology_cols = ontologies

    else:
        # Auto-detect ontology columns by matching against ONTOLOGIES_LIST
        ontology_cols = set(wide_df.columns) & set(ONTOLOGIES_LIST)
        if not ontology_cols:
            raise ValueError(
                f"No valid ontology columns found in DataFrame. Column names must match one of: {ONTOLOGIES_LIST}"
            )
        logger.info(f"Auto-detected ontology columns: {ontology_cols}")

    logger.debug(f"Validated ontology columns: {ontology_cols}")
    return ontology_cols


def _ensure_feature_id_var(
    df: pd.DataFrame, feature_id_var: str = FEATURE_ID_VAR_DEFAULT
) -> pd.DataFrame:
    """
    Ensure the DataFrame has a feature_id column, creating one if it doesn't exist.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to check/modify
    feature_id_var : str, default=FEATURE_ID_VAR_DEFAULT
        Name of the feature ID column

    Returns
    -------
    pd.DataFrame
        DataFrame with guaranteed feature_id column
    """
    if feature_id_var not in df.columns:
        logger.warning(f"No {feature_id_var} column found in DataFrame, creating one")
        df = df.copy()
        df[feature_id_var] = np.arange(len(df))
    return df


def _get_numeric_aggregator(
    method: str = RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN,
    feature_id_var: str = FEATURE_ID_VAR_DEFAULT,
) -> callable:
    """
    Get aggregation function for numeric columns with various methods.

    Parameters
    ----------
    method : str, default="weighted_mean"
        Aggregation method to use:
        - "weighted_mean": weighted by inverse of feature_id frequency (default)
        - "mean": simple arithmetic mean
        - "first": first value after sorting by feature_id_var (requires feature_id_var)
        - "max": maximum value
    feature_id_var : str, default="feature_id"
        Name of the column specifying a measured feature - used for sorting and weighting

    Returns
    -------
    callable
        Aggregation function to use with groupby

    Raises
    ------
    ValueError
        If method is not recognized
    """

    def weighted_mean(df: pd.DataFrame) -> float:
        # Get values and weights for this group
        values = df["value"]
        weights = df["weight"]
        # Weights are already normalized globally, just use them directly
        return (values * weights).sum() / weights.sum()

    def first_by_id(df: pd.DataFrame) -> float:
        # Sort by feature_id and take first value
        return df.sort_values(feature_id_var).iloc[0]["value"]

    def simple_mean(series: pd.Series) -> float:
        return series.mean()

    def simple_max(series: pd.Series) -> float:
        return series.max()

    aggregators = {
        RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN: weighted_mean,
        RESOLVE_MATCHES_AGGREGATORS.MEAN: simple_mean,
        RESOLVE_MATCHES_AGGREGATORS.FIRST: first_by_id,
        RESOLVE_MATCHES_AGGREGATORS.MAX: simple_max,
    }

    if method not in aggregators:
        raise ValueError(
            f"Unknown aggregation method: {method}. Must be one of {list(aggregators.keys())}"
        )

    return aggregators[method]


def _split_numeric_non_numeric_columns(df: pd.DataFrame, always_non_numeric=None):
    """
    Utility to split DataFrame columns into numeric and non-numeric, always treating specified columns as non-numeric.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to split.
    always_non_numeric : list or set, optional
        Columns to always treat as non-numeric (e.g., ['feature_id']).

    Returns
    -------
    numeric_cols : pd.Index
        Columns considered numeric (int64, float64, and not in always_non_numeric).
    non_numeric_cols : pd.Index
        Columns considered non-numeric (object, string, etc., plus always_non_numeric).
    """
    if always_non_numeric is None:
        always_non_numeric = []
    always_non_numeric = set(always_non_numeric)
    numeric_cols = df.select_dtypes(include=["int64", "float64"]).columns.difference(
        always_non_numeric
    )
    non_numeric_cols = df.columns.difference(numeric_cols)
    return numeric_cols, non_numeric_cols


def _aggregate_grouped_columns(
    df: pd.DataFrame,
    numeric_cols,
    non_numeric_cols,
    numeric_aggregator,
    feature_id_var: str = FEATURE_ID_VAR_DEFAULT,
    numeric_agg: str = RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN,
) -> pd.DataFrame:
    """
    Aggregate numeric and non-numeric columns for grouped DataFrame.
    Assumes deduplication by feature_id within each s_id has already been performed.
    Returns the combined DataFrame.
    """
    results = []

    # Handle non-numeric columns
    if len(non_numeric_cols) > 0:
        non_numeric_agg = (
            df[non_numeric_cols]
            .groupby(level=0)
            .agg(lambda x: ",".join(sorted(set(x.astype(str)))))
        )
        results.append(non_numeric_agg)
    # Handle numeric columns
    if len(numeric_cols) > 0:
        numeric_results = {}
        for col in numeric_cols:
            if numeric_agg in [
                RESOLVE_MATCHES_AGGREGATORS.FIRST,
                RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN,
            ]:
                agg_df = pd.DataFrame(
                    {"value": df[col], feature_id_var: df[feature_id_var]}
                )
                if numeric_agg == RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN:
                    agg_df[RESOLVE_MATCHES_TMP_WEIGHT_COL] = df[
                        RESOLVE_MATCHES_TMP_WEIGHT_COL
                    ]
                numeric_results[col] = agg_df.groupby(level=0).apply(
                    lambda x: (
                        numeric_aggregator(x)
                        if numeric_agg != RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN
                        else numeric_aggregator(
                            x.rename(columns={RESOLVE_MATCHES_TMP_WEIGHT_COL: "weight"})
                        )
                    )
                )
            else:
                numeric_results[col] = df[col].groupby(level=0).agg(numeric_aggregator)
        numeric_agg_df = pd.DataFrame(numeric_results)
        results.append(numeric_agg_df)
    # Combine results
    if results:
        resolved = pd.concat(results, axis=1)
    else:
        resolved = pd.DataFrame(index=df.index)
    return resolved


def _log_feature_species_mapping_stats(
    pathway_species: pd.DataFrame, feature_id_var: str = FEATURE_ID_VAR_DEFAULT
):
    """
    Log statistics about the mapping between feature_id and s_id in the pathway_species DataFrame.
    """

    # Percent of feature_ids present one or more times in the output
    n_feature_ids = pathway_species[feature_id_var].nunique()
    n_input_feature_ids = (
        pathway_species[feature_id_var].max() + 1
        if feature_id_var in pathway_species.columns
        else 0
    )
    percent_present = (
        100 * n_feature_ids / n_input_feature_ids if n_input_feature_ids else 0
    )
    logger.info(
        f"{percent_present:.1f}% of feature_ids are present one or more times in the output ({n_feature_ids}/{n_input_feature_ids})"
    )

    # Number of times an s_id maps to 1+ feature_ids (with s_name)
    s_id_counts = pathway_species.groupby(SBML_DFS.S_ID)[feature_id_var].nunique()
    s_id_multi = s_id_counts[s_id_counts > 1]
    logger.info(f"{len(s_id_multi)} s_id(s) map to more than one feature_id.")
    if not s_id_multi.empty:
        examples = pathway_species[
            pathway_species[SBML_DFS.S_ID].isin(s_id_multi.index)
        ][[SBML_DFS.S_ID, SBML_DFS.S_NAME, feature_id_var]]
        logger.info(
            f"Examples of s_id mapping to multiple feature_ids (showing up to 3):\n{examples.groupby([SBML_DFS.S_ID, SBML_DFS.S_NAME])[feature_id_var].apply(list).head(3)}"
        )

    # Number of times a feature_id maps to 1+ s_ids (with s_name)
    feature_id_counts = pathway_species.groupby(feature_id_var)[SBML_DFS.S_ID].nunique()
    feature_id_multi = feature_id_counts[feature_id_counts > 1]
    logger.info(f"{len(feature_id_multi)} feature_id(s) map to more than one s_id.")
    if not feature_id_multi.empty:
        examples = pathway_species[
            pathway_species[feature_id_var].isin(feature_id_multi.index)
        ][[feature_id_var, SBML_DFS.S_ID, SBML_DFS.S_NAME]]
        logger.info(
            f"Examples of feature_id mapping to multiple s_ids (showing up to 3):\n{examples.groupby([feature_id_var])[[SBML_DFS.S_ID, SBML_DFS.S_NAME]].apply(lambda df: list(df.itertuples(index=False, name=None))).head(3)}"
        )



"""
Function execution components for the Napistu MCP server.
"""

from typing import Dict, List, Any, Optional
import inspect

# Global storage for session context and objects
_session_context = {}
_session_objects = {}


async def initialize_components() -> bool:
    """
    Initialize execution components.

    Returns
    -------
    bool
        True if initialization is successful.
    """
    global _session_context, _session_objects
    import napistu

    _session_context["napistu"] = napistu
    return True


def register_object(name: str, obj: Any) -> None:
    """
    Register an object with the execution component.

    Args:
        name: Name to reference the object by
        obj: The object to register
    """
    global _session_objects
    _session_objects[name] = obj
    print(f"Registered object '{name}' with MCP server")


def register_components(mcp, session_context=None, object_registry=None):
    """
    Register function execution components with the MCP server.

    Args:
        mcp: FastMCP server instance
        session_context: Dictionary of the user's current session (e.g., globals())
        object_registry: Dictionary of named objects to make available
    """
    global _session_context, _session_objects

    # Initialize context
    if session_context:
        _session_context = session_context

    if object_registry:
        _session_objects = object_registry

    import napistu

    _session_context["napistu"] = napistu

    # Register resources
    @mcp.resource("napistu-local://registry")
    async def get_registry_summary() -> Dict[str, Any]:
        """
        Get a summary of all objects registered with the server.
        """
        return {
            "object_count": len(_session_objects),
            "object_names": list(_session_objects.keys()),
            "object_types": {
                name: type(obj).__name__ for name, obj in _session_objects.items()
            },
        }

    @mcp.resource("napistu-local://environment")
    async def get_environment_info() -> Dict[str, Any]:
        """
        Get information about the local Python environment.
        """
        try:
            import napistu

            napistu_version = getattr(napistu, "__version__", "unknown")
        except ImportError:
            napistu_version = "not installed"

        import sys

        return {
            "python_version": sys.version,
            "napistu_version": napistu_version,
            "platform": sys.platform,
            "registered_objects": list(_session_objects.keys()),
        }

    # Register tools
    @mcp.tool()
    async def list_registry() -> Dict[str, Any]:
        """
        List all objects registered with the server.

        Returns:
            Dictionary with information about registered objects
        """
        result = {}

        for name, obj in _session_objects.items():
            obj_type = type(obj).__name__

            # Get additional info based on object type
            if hasattr(obj, "shape"):  # For pandas DataFrame or numpy array
                obj_info = {
                    "type": obj_type,
                    "shape": str(obj.shape),
                }
            elif hasattr(obj, "__len__"):  # For lists, dicts, etc.
                obj_info = {
                    "type": obj_type,
                    "length": len(obj),
                }
            else:
                obj_info = {
                    "type": obj_type,
                }

            result[name] = obj_info

        return result

    @mcp.tool()
    async def describe_object(object_name: str) -> Dict[str, Any]:
        """
        Get detailed information about a registered object.

        Args:
            object_name: Name of the registered object

        Returns:
            Dictionary with object information
        """
        if object_name not in _session_objects:
            return {"error": f"Object '{object_name}' not found in registry"}

        obj = _session_objects[object_name]
        obj_type = type(obj).__name__

        # Basic info for all objects
        result = {
            "name": object_name,
            "type": obj_type,
            "methods": [],
            "attributes": [],
        }

        # Add methods and attributes
        for name in dir(obj):
            if name.startswith("_"):
                continue

            try:
                attr = getattr(obj, name)

                if callable(attr):
                    # Method
                    sig = str(inspect.signature(attr))
                    doc = inspect.getdoc(attr) or ""
                    result["methods"].append(
                        {
                            "name": name,
                            "signature": sig,
                            "docstring": doc,
                        }
                    )
                else:
                    # Attribute
                    attr_type = type(attr).__name__
                    result["attributes"].append(
                        {
                            "name": name,
                            "type": attr_type,
                        }
                    )
            except Exception:
                # Skip attributes that can't be accessed
                pass

        return result

    @mcp.tool()
    async def execute_function(
        function_name: str,
        object_name: Optional[str] = None,
        args: Optional[List] = None,
        kwargs: Optional[Dict] = None,
    ) -> Dict[str, Any]:
        """
        Execute a Napistu function on a registered object.

        Args:
            function_name: Name of the function to execute
            object_name: Name of the registered object to operate on (if method call)
            args: Positional arguments to pass to the function
            kwargs: Keyword arguments to pass to the function

        Returns:
            Dictionary with execution results
        """
        args = args or []
        kwargs = kwargs or {}

        try:
            if object_name:
                # Method call on an object
                if object_name not in _session_objects:
                    return {"error": f"Object '{object_name}' not found in registry"}

                obj = _session_objects[object_name]

                if not hasattr(obj, function_name):
                    return {
                        "error": f"Method '{function_name}' not found on object '{object_name}'"
                    }

                func = getattr(obj, function_name)
                result = func(*args, **kwargs)
            else:
                # Global function call
                if function_name in _session_context:
                    # Function from session context
                    func = _session_context[function_name]
                    result = func(*args, **kwargs)
                else:
                    # Try to find the function in Napistu
                    try:
                        import napistu

                        # Split function name by dots for nested modules
                        parts = function_name.split(".")
                        current = napistu

                        for part in parts[:-1]:
                            current = getattr(current, part)

                        func = getattr(current, parts[-1])
                        result = func(*args, **kwargs)
                    except (ImportError, AttributeError):
                        return {"error": f"Function '{function_name}' not found"}

            # Register result if it's a return value
            if result is not None:
                result_name = f"result_{len(_session_objects) + 1}"
                _session_objects[result_name] = result

                # Basic type conversion for JSON serialization
                if hasattr(result, "to_dict"):
                    # For pandas DataFrame or similar
                    return {
                        "success": True,
                        "result_name": result_name,
                        "result_type": type(result).__name__,
                        "result_preview": (
                            result.to_dict()
                            if hasattr(result, "__len__") and len(result) < 10
                            else "Result too large to preview"
                        ),
                    }
                elif hasattr(result, "to_json"):
                    # For objects with JSON serialization
                    return {
                        "success": True,
                        "result_name": result_name,
                        "result_type": type(result).__name__,
                        "result_preview": result.to_json(),
                    }
                elif hasattr(result, "__dict__"):
                    # For custom objects
                    return {
                        "success": True,
                        "result_name": result_name,
                        "result_type": type(result).__name__,
                        "result_preview": str(result),
                    }
                else:
                    # For simple types
                    return {
                        "success": True,
                        "result_name": result_name,
                        "result_type": type(result).__name__,
                        "result_preview": str(result),
                    }
            else:
                return {
                    "success": True,
                    "result": None,
                }
        except Exception as e:
            import traceback

            return {
                "error": str(e),
                "traceback": traceback.format_exc(),
            }

    @mcp.tool()
    async def search_paths(
        source_node: str,
        target_node: str,
        network_object: str,
        max_depth: int = 3,
    ) -> Dict[str, Any]:
        """
        Find paths between two nodes in a network.

        Args:
            source_node: Source node identifier
            target_node: Target node identifier
            network_object: Name of the registered network object
            max_depth: Maximum path length

        Returns:
            Dictionary with paths found
        """
        if network_object not in _session_objects:
            return {"error": f"Network object '{network_object}' not found in registry"}

        network = _session_objects[network_object]

        try:
            # Import necessary modules
            import napistu

            # Check if the object is a valid network type
            if hasattr(network, "find_paths"):
                # Direct method call
                paths = network.find_paths(
                    source_node, target_node, max_depth=max_depth
                )
            elif hasattr(napistu.graph, "find_paths"):
                # Function call
                paths = napistu.graph.find_paths(
                    network, source_node, target_node, max_depth=max_depth
                )
            else:
                return {"error": "Could not find appropriate path-finding function"}

            # Register result
            result_name = f"paths_{len(_session_objects) + 1}"
            _session_objects[result_name] = paths

            # Return results
            if hasattr(paths, "to_dict"):
                return {
                    "success": True,
                    "result_name": result_name,
                    "paths_found": (
                        len(paths) if hasattr(paths, "__len__") else "unknown"
                    ),
                    "result_preview": (
                        paths.to_dict()
                        if hasattr(paths, "__len__") and len(paths) < 10
                        else "Result too large to preview"
                    ),
                }
            else:
                return {
                    "success": True,
                    "result_name": result_name,
                    "paths_found": (
                        len(paths) if hasattr(paths, "__len__") else "unknown"
                    ),
                    "result_preview": str(paths),
                }
        except Exception as e:
            import traceback

            return {
                "error": str(e),
                "traceback": traceback.format_exc(),
            }
